# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

from plover.orthography import add_suffix

from plover_build_utils.testing import parametrize


ADD_SUFFIX_TESTS = (
    lambda: ("artistic", "ly", "artistically"),
    lambda: ("cosmetic", "ly", "cosmetically"),
    lambda: ("establish", "s", "establishes"),
    lambda: ("speech", "s", "speeches"),
    lambda: ("approach", "s", "approaches"),
    lambda: ("beach", "s", "beaches"),
    lambda: ("arch", "s", "arches"),
    lambda: ("larch", "s", "larches"),
    lambda: ("march", "s", "marches"),
    lambda: ("search", "s", "searches"),
    lambda: ("starch", "s", "starches"),
    lambda: ("stomach", "s", "stomachs"),
    lambda: ("monarch", "s", "monarchs"),
    lambda: ("patriarch", "s", "patriarchs"),
    lambda: ("oligarch", "s", "oligarchs"),
    lambda: ("cherry", "s", "cherries"),
    lambda: ("day", "s", "days"),
    lambda: ("penny", "s", "pennies"),
    lambda: ("pharmacy", "ist", "pharmacist"),
    lambda: ("melody", "ist", "melodist"),
    lambda: ("pacify", "ist", "pacifist"),
    lambda: ("geology", "ist", "geologist"),
    lambda: ("metallurgy", "ist", "metallurgist"),
    lambda: ("anarchy", "ist", "anarchist"),
    lambda: ("monopoly", "ist", "monopolist"),
    lambda: ("alchemy", "ist", "alchemist"),
    lambda: ("botany", "ist", "botanist"),
    lambda: ("therapy", "ist", "therapist"),
    lambda: ("theory", "ist", "theorist"),
    lambda: ("psychiatry", "ist", "psychiatrist"),
    lambda: ("lobby", "ist", "lobbyist"),
    lambda: ("hobby", "ist", "hobbyist"),
    lambda: ("copy", "ist", "copyist"),
    lambda: ("beauty", "ful", "beautiful"),
    lambda: ("pretty", "ify", "prettify"),
    lambda: ("weary", "ness", "weariness"),
    lambda: ("weary", "some", "wearisome"),
    lambda: ("lonely", "ness", "loneliness"),
    lambda: ("narrate", "ing", "narrating"),
    lambda: ("narrate", "or", "narrator"),
    lambda: ("generalize", "ability", "generalizability"),
    lambda: ("reproduce", "able", "reproducible"),
    lambda: ("grade", "ations", "gradations"),
    lambda: ("urine", "ary", "urinary"),
    lambda: ("achieve", "able", "achievable"),
    lambda: ("polarize", "ation", "polarization"),
    lambda: ("done", "or", "donor"),
    lambda: ("analyze", "ed", "analyzed"),
    lambda: ("narrate", "ing", "narrating"),
    lambda: ("believe", "able", "believable"),
    lambda: ("animate", "ors", "animators"),
    lambda: ("discontinue", "ation", "discontinuation"),
    lambda: ("innovate", "ive", "innovative"),
    lambda: ("future", "ists", "futurists"),
    lambda: ("illustrate", "or", "illustrator"),
    lambda: ("emerge", "ent", "emergent"),
    lambda: ("equip", "ed", "equipped"),
    lambda: ("defer", "ed", "deferred"),
    lambda: ("defer", "er", "deferrer"),
    lambda: ("defer", "ing", "deferring"),
    lambda: ("pigment", "ed", "pigmented"),
    lambda: ("refer", "ed", "referred"),
    lambda: ("fix", "ed", "fixed"),
    lambda: ("alter", "ed", "altered"),
    lambda: ("interpret", "ing", "interpreting"),
    lambda: ("wonder", "ing", "wondering"),
    lambda: ("target", "ing", "targeting"),
    lambda: ("limit", "er", "limiter"),
    lambda: ("maneuver", "ing", "maneuvering"),
    lambda: ("monitor", "ing", "monitoring"),
    lambda: ("color", "ing", "coloring"),
    lambda: ("inhibit", "ing", "inhibiting"),
    lambda: ("master", "ed", "mastered"),
    lambda: ("target", "ing", "targeting"),
    lambda: ("fix", "ed", "fixed"),
    lambda: ("scrap", "y", "scrappy"),
    lambda: ("trip", "s", "trips"),
    lambda: ("equip", "s", "equips"),
    lambda: ("bat", "en", "batten"),
    lambda: ("smite", "en", "smitten"),
    lambda: ("got", "en", "gotten"),
    lambda: ("bite", "en", "bitten"),
    lambda: ("write", "en", "written"),
    lambda: ("flax", "en", "flaxen"),
    lambda: ("wax", "en", "waxen"),
    lambda: ("fast", "est", "fastest"),
    lambda: ("white", "er", "whiter"),
    lambda: ("crap", "y", "crappy"),
    lambda: ("lad", "er", "ladder"),
    lambda: ("translucent", "cy", "translucency"),
    lambda: ("bankrupt", "cy", "bankruptcy"),
    lambda: ("inadequate", "cy", "inadequacy"),
    lambda: ("secret", "cy", "secrecy"),
    lambda: ("impolite", "cy", "impolicy"),
    lambda: ("idiot", "cy", "idiocy"),
    lambda: ("free", "ed", "freed"),
    lambda: ("free", "er", "freer"),
    lambda: ("regulate", "ry", "regulatory"),
    lambda: ("humble", "ly", "humbly"),
    lambda: ("questionable", "ly", "questionably"),
    lambda: ("triple", "ly", "triply"),
    lambda: ("statute", "ry", "statutory"),
    lambda: ("statute", "ary", "statutory"),
    lambda: ("confirm", "tory", "confirmatory"),
    lambda: ("supervise", "ary", "supervisory"),
    lambda: ("Minessota", "en", "Minessotan"),
    lambda: ("ceremony", "ial", "ceremonial"),
    lambda: ("editor", "ial", "editorial"),
    lambda: ("editor", "ially", "editorially"),
    lambda: ("spaghetti", "ification", "spaghettification"),
    lambda: ("Hawaii", "ification", "Hawaiification"),
    lambda: ("ASCII", "ification", "ASCIIfication"),
    lambda: ("play", "ification", "playification"),
    lambda: ("fantastic", "ical", "fantastical"),
    lambda: ("epistomology", "ical", "epistomological"),
    lambda: ("oratory", "ical", "oratorical"),
    lambda: ("radical", "ist", "radicalist"),
    lambda: ("complementary", "ity", "complementarity"),
    lambda: ("disproportional", "ity", "disproportionality"),
    lambda: ("perform", "tive", "performative"),
    lambda: ("restore", "tive", "restorative"),
    lambda: ("token", "ize", "tokenize"),
    lambda: ("token", "ise", "tokenise"),
    lambda: ("conditional", "ize", "conditionalize"),
    lambda: ("conditional", "isation", "conditionalisation"),
    lambda: ("spectacular", "ization", "spectacularization"),
    lambda: ("spectacular", "ism", "spectacularism"),
    lambda: ("category", "ize", "categorize"),
    lambda: ("category", "ise", "categorise"),
    lambda: ("custom", "izable", "customizable"),
    lambda: ("custom", "isable", "customisable"),
    lambda: ("fantasy", "ize", "fantasize"),
    lambda: ("criminal", "ology", "criminology"),
    lambda: ("criminal", "ologist", "criminologist"),
    lambda: ("dermatology", "ist", "dermatologist"),
    lambda: ("similar", "ish", "similarish"),
)


@parametrize(ADD_SUFFIX_TESTS)
def test_add_suffix(word, suffix, expected):
    assert add_suffix(word, suffix) == expected



# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.
#
# TODO: Convert non-ascii characters to UTF8
# TODO: What does ^ mean in Eclipse?
# TODO: What does #N mean in Eclipse?
# TODO: convert supported commands from Eclipse

"""Parsing an RTF/CRE dictionary.

RTF/CRE spec:
https://web.archive.org/web/20201017075356/http://www.legalxml.org/workgroups/substantive/transcripts/cre-spec.htm

"""

import re
import string

from plover import __version__ as plover_version
from plover.dictionary.helpers import StenoNormalizer
from plover.formatting import ATOM_RE
from plover.steno_dictionary import StenoDictionary

from .rtfcre_parse import parse_rtfcre


HEADER = (
    r"{\rtf1\ansi{\*\cxrev100}"
    r"\cxdict{\*\cxsystem Plover %s}"
    r"{\stylesheet{\s0 Normal;}}"
) % plover_version


class RegexFormatter:
    def __init__(self, spec_list, escape_fn):
        self._escape_fn = escape_fn
        self._format_for_lastindex = [None]
        pattern_list = []
        for pattern, replacement in spec_list:
            num_groups = len(self._format_for_lastindex)
            pattern_groups = re.compile(pattern).groups
            if pattern_groups:
                needed = []
                for token in string.Formatter().parse(replacement):
                    field_name = token[1]
                    if not field_name:
                        continue
                    group = int(field_name)
                    assert 0 <= group <= pattern_groups
                    needed.append(group + num_groups)
            else:
                pattern = "(" + pattern + ")"
                pattern_groups = 1
                needed = []
            for n in range(pattern_groups):
                self._format_for_lastindex.append((needed, replacement))
            pattern_list.append(pattern)
        self._format_rx = re.compile("|".join(pattern_list))

    def format(self, s):
        m = self._format_rx.fullmatch(s)
        if m is None:
            return None
        needed, replacement = self._format_for_lastindex[m.lastindex]
        return replacement.format(*(self._escape_fn(m.group(g)) for g in needed))


class TranslationFormatter:
    # fmt: off
    TO_ESCAPE = (
        (r"([\\{}])", r"\\\1"   ),
        (r"\n\n"    , r"\\par " ),
        (r"\n"      , r"\\line "),
        (r"\t"      , r"\\tab " ),
    )
    ATOMS_FORMATTERS = (
        # Note order matters!
        (r"{\.}"                       , r"{{\cxp. }}"                 ),
        (r"{!}"                        , r"{{\cxp! }}"                 ),
        (r"{\?}"                       , r"{{\cxp? }}"                 ),
        (r"{\,}"                       , r"{{\cxp, }}"                 ),
        (r"{:}"                        , r"{{\cxp: }}"                 ),
        (r"{;}"                        , r"{{\cxp; }}"                 ),
        (r"{\^ \^}"                    , r"\~"                         ),
        (r"{\^-\^}"                    , r"\_"                         ),
        (r"{\^\^?}"                    , r"{{\cxds}}"                  ),
        (r"{\^([^^}]*)\^}"             , r"{{\cxds {0}\cxds}}"         ),
        (r"{\^([^^}]*)}"               , r"{{\cxds {0}}}"              ),
        (r"{([^^}]*)\^}"               , r"{{{0}\cxds}}"               ),
        (r"{-\|}"                      , r"\cxfc "                     ),
        (r"{>}"                        , r"\cxfl "                     ),
        (r"{ }"                        , r" "                          ),
        (r"{&([^}]+)}"                 , r"{{\cxfing {0}}}"            ),
        (r"{(.*)}"                     , r"{{\*\cxplovermeta {0}}}"    ),
    )
    TRANSLATIONS_FORMATTERS = (
        (r"{\*}"                       , r"{{\*\cxplovermacro retrospective_toggle_asterisk}}"),
        (r"{\*!}"                      , r"{{\*\cxplovermacro retrospective_delete_space}}"),
        (r"{\*\?}"                     , r"{{\*\cxplovermacro retrospective_insert_space}}"),
        (r"{\*\+}"                     , r"{{\*\cxplovermacro repeat_last_stroke}}"),
        (r"=undo"                      , r"\cxdstroke"                 ),
        (r"=(\w+(?::.*)?)"             , r"{{\*\cxplovermacro {0}}}"   ),
    )
    # fmt: on

    def __init__(self):
        self._to_escape = [
            (re.compile(pattern), replacement)
            for pattern, replacement in self.TO_ESCAPE
        ]
        self._atom_formatter = RegexFormatter(self.ATOMS_FORMATTERS, self.escape)
        self._translation_formatter = RegexFormatter(
            self.TRANSLATIONS_FORMATTERS, self.escape
        )

    def escape(self, text):
        for rx, replacement in self._to_escape:
            text = rx.sub(replacement, text)
        return text

    def format(self, translation):
        s = self._translation_formatter.format(translation)
        if s is not None:
            return s
        parts = []
        for atom in ATOM_RE.findall(translation):
            atom = atom.strip()
            if not atom:
                continue
            s = self._atom_formatter.format(atom)
            if s is None:
                s = self.escape(atom)
            parts.append(s)
        return "".join(parts)


class RtfDictionary(StenoDictionary):
    def _load(self, filename):
        with open(filename, "rb") as fp:
            text = fp.read().decode("cp1252")
        with StenoNormalizer(filename) as normalize_steno:
            self.update(parse_rtfcre(text, normalize=normalize_steno))

    def _save(self, filename):
        translation_formatter = TranslationFormatter()
        with open(filename, "w", encoding="cp1252", newline="\r\n") as fp:
            print(HEADER, file=fp)
            for s, t in self.items():
                s = "/".join(s)
                t = translation_formatter.format(t)
                entry = r"{\*\cxs %s}%s" % (s, t)
                print(entry, file=fp)
            print("}", file=fp)



import os
import signal
import subprocess
import sys
import threading

from PySide6.QtCore import (
    Signal,
)
from PySide6.QtGui import QFontDatabase, QFontMetrics
from PySide6.QtWidgets import QWidget

from plover.gui_qt.console_widget_ui import Ui_ConsoleWidget


NULL = open(os.devnull, "r+b")


class ConsoleWidget(QWidget, Ui_ConsoleWidget):
    textOutput = Signal(str)
    processFinished = Signal(object)

    def __init__(self, popen=None):
        super().__init__()
        self.setupUi(self)
        self.textOutput.connect(self.output.append)
        self._popen = subprocess.Popen if popen is None else popen
        self._proc = None
        self._thread = None
        font = QFontDatabase.systemFont(QFontDatabase.SystemFont.FixedFont)
        metrics = QFontMetrics(font)
        self.output.setMinimumSize(80 * metrics.maxWidth(), 24 * metrics.height())
        self.output.setCurrentFont(font)

    def run(self, args):
        assert self._thread is None
        if sys.platform.startswith("win32"):
            # Make it possible to interrupt by sending a Ctrl+C event.
            kwargs = {"creationflags": subprocess.CREATE_NEW_PROCESS_GROUP}
        else:
            kwargs = {}
        self._proc = self._popen(
            args, stdin=NULL, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, **kwargs
        )
        self._thread = threading.Thread(target=self._subprocess)
        self._thread.start()

    def terminate(self):
        assert self._proc is not None
        if sys.platform.startswith("win32"):
            sig = signal.CTRL_C_EVENT
        else:
            sig = signal.SIGINT
        self._proc.send_signal(sig)
        try:
            self._proc.wait(10)
        except subprocess.TimeoutExpired:
            self._proc.terminate()
        self._thread.join()

    def _subprocess(self):
        while True:
            try:
                line = self._proc.stdout.readline()
            except:
                break
            if not line:
                break
            line = line.decode()
            if line.endswith(os.linesep):
                line = line[: -len(os.linesep)]
            print(line)
            self.textOutput.emit(line)
        self.processFinished.emit(self._proc.wait())



# Copyright (c) 2012 Hesky Fisher
# See LICENSE.txt for details.

"""Platform dependent configuration."""

import os
import sys

import appdirs


if sys.platform.startswith("darwin"):
    PLATFORM = "mac"
elif sys.platform.startswith("linux"):
    PLATFORM = "linux"
elif sys.platform.startswith("win"):
    PLATFORM = "win"
elif sys.platform.startswith(("freebsd", "openbsd")):
    PLATFORM = "bsd"
else:
    PLATFORM = None

# If the program's working directory has a plover.cfg file then run in
# "portable mode", i.e. store all data in the same directory. This allows
# keeping all Plover files in a portable drive.
#
# Note: the special case when run from an app bundle on macOS.
if PLATFORM == "mac" and ".app/" in os.path.realpath(__file__):
    PROGRAM_DIR = os.path.abspath(
        os.path.join(os.path.dirname(sys.executable), *[os.path.pardir] * 3)
    )
else:
    PROGRAM_DIR = os.getcwd()

# Setup configuration directory.
CONFIG_BASENAME = "plover.cfg"
if os.path.isfile(os.path.join(PROGRAM_DIR, CONFIG_BASENAME)):
    CONFIG_DIR = PROGRAM_DIR
else:
    config_directories = [
        getattr(appdirs, directory_type)("plover")
        for directory_type in ("user_config_dir", "user_data_dir")
    ]
    for CONFIG_DIR in config_directories:
        if os.path.isfile(os.path.join(CONFIG_DIR, CONFIG_BASENAME)):
            break
    else:
        CONFIG_DIR = config_directories[0]
CONFIG_FILE = os.path.join(CONFIG_DIR, CONFIG_BASENAME)

# Setup plugins directory.
PLUGINS_PLATFORM = PLATFORM

ASSETS_DIR = os.path.realpath(os.path.join(__file__, "../../assets"))



from PySide6.QtCore import QEvent, Qt, Slot

from plover import _
from plover.translation import unescape_translation

from plover.gui_qt.lookup_dialog_ui import Ui_LookupDialog
from plover.gui_qt.tool import Tool


class LookupDialog(Tool, Ui_LookupDialog):
    # i18n: Widget: “LookupDialog”, tooltip.
    __doc__ = _("Search the dictionary for translations.")

    TITLE = _("Lookup")
    ICON = ":/resources/lookup.svg"
    ROLE = "lookup"
    SHORTCUT = "Ctrl+L"

    def __init__(self, engine):
        super().__init__(engine)
        self.setupUi(self)
        self.pattern.installEventFilter(self)
        self.suggestions.installEventFilter(self)
        self.pattern.setFocus()
        self.restore_state()
        self.finished.connect(self.save_state)

    def eventFilter(self, watched, event):
        if event.type() == QEvent.Type.KeyPress and event.key() in (
            Qt.Key.Key_Enter,
            Qt.Key.Key_Return,
        ):
            return True
        return False

    def _update_suggestions(self, suggestion_list):
        self.suggestions.clear()
        self.suggestions.append(suggestion_list)

    @Slot(str)
    def lookup(self, pattern):
        translation = unescape_translation(pattern.strip())
        suggestion_list = self._engine.get_suggestions(translation)
        self._update_suggestions(suggestion_list)

    def changeEvent(self, event):
        super().changeEvent(event)
        if event.type() == QEvent.Type.ActivationChange and self.isActiveWindow():
            self.pattern.setFocus()
            self.pattern.selectAll()



# Copyright (c) 2010-2011 Joshua Harlan Lifton.
# See LICENSE.txt for details.

"""This module converts translations to printable text.

This module defines and implements plover's custom dictionary language.

"""

from enum import Enum
from os.path import commonprefix
from collections import namedtuple
import re
import string

from plover.registry import registry


Case = Enum(
    "case",
    (
        (c, c.lower())
        for c in """
                     CAP_FIRST_WORD
                     LOWER
                     LOWER_FIRST_CHAR
                     TITLE
                     UPPER
                     UPPER_FIRST_WORD
                     """.split()
    ),
)

SPACE = " "

META_ATTACH_FLAG = "^"
META_CARRY_CAPITALIZATION = "~|"
META_GLUE_FLAG = "&"

META_ESCAPE = "\\"
RE_META_ESCAPE = "\\\\"
META_START = "{"
META_END = "}"
META_ESC_START = META_ESCAPE + META_START
META_ESC_END = META_ESCAPE + META_END


def _build_metas_parser(supported_metas):
    match_result_from_lastindex = [None]
    regex_parts = []

    for pattern, name, param in supported_metas:
        num_prev_groups = len(match_result_from_lastindex)
        if isinstance(pattern, tuple):
            pattern = pattern[0].format(
                *["(?:" + re.escape(s) + ")" for s in pattern[1:]]
            )
        num_groups = re.compile(pattern).groups
        if isinstance(name, int):
            assert num_groups > 0
            assert 0 <= name < num_groups
            name += num_prev_groups
        if isinstance(param, int):
            assert num_groups > 0
            assert 0 <= param < num_groups
            param += num_prev_groups
        if num_groups == 0:
            num_groups = 1
        else:
            pattern = "?:" + pattern
        groups = [n + num_prev_groups for n in range(num_groups)]
        match_result_from_lastindex.extend(((name, param),) * len(groups))
        regex_parts.append("(" + pattern + ")$")

    regex = re.compile("|".join(regex_parts), re.DOTALL | re.IGNORECASE)

    def parse(meta):
        m = regex.match(meta)
        if m is None:
            return None, None
        meta_name, meta_param = match_result_from_lastindex[m.lastindex]
        if isinstance(meta_name, int):
            meta_name = m.group(meta_name)
        if isinstance(meta_param, int):
            meta_param = m.group(meta_param)
        return [meta_name, meta_param]

    return parse


# Note: declaration order matters!
_parse_meta = _build_metas_parser(
    (
        # Generic {:macro:cmdline} syntax.
        (r":([^:]+):?(.*)", 0, 1),
        # Command.
        (r"PLOVER:(.*)", "command", 0),
        # Key combination.
        (r"#(.*)", "key_combo", 0),
        # Punctuation.
        (r"([,:;])", "comma", 0),
        (r"([.!?])", "stop", 0),
        # Case.
        (r"-\|", "case", Case.CAP_FIRST_WORD.value),
        (r">", "case", Case.LOWER_FIRST_CHAR.value),
        (r"<", "case", Case.UPPER_FIRST_WORD.value),
        (r"\*-\|", "retro_case", Case.CAP_FIRST_WORD.value),
        (r"\*>", "retro_case", Case.LOWER_FIRST_CHAR.value),
        (r"\*<", "retro_case", Case.UPPER_FIRST_WORD.value),
        # Explicit word end.
        (r"(\$)", "word_end", 0),
        # Conditional.
        (r"=(.*)", "if_next_matches", 0),
        # Mode.
        (r"MODE:(.*)", "mode", 0),
        # Currency.
        (r"\*\((.*)\)", "retro_currency", 0),
        # Glue.
        ((r"{0}(.*)", META_GLUE_FLAG), "glue", 0),
        # Carry capitalization.
        (
            (r"({0}?{1}.*{0}?)", META_ATTACH_FLAG, META_CARRY_CAPITALIZATION),
            "carry_capitalize",
            0,
        ),
        # Attach.
        ((r"({0}.*{0}?)", META_ATTACH_FLAG), "attach", 0),
        ((r"(.*{0})", META_ATTACH_FLAG), "attach", 0),
    )
)


ATOM_RE = re.compile(
    r"""(?:%s%s|%s%s|[^%s%s])+ # One or more of anything
                                                # other than unescaped { or }
                                                #
                                              | # or
                                                #
                     %s(?:%s%s|%s%s|[^%s%s])*%s # Anything of the form {X}
                                                # where X doesn't contain
                                                # unescaped { or }
                      """
    % (
        RE_META_ESCAPE,
        META_START,
        RE_META_ESCAPE,
        META_END,
        META_START,
        META_END,
        META_START,
        RE_META_ESCAPE,
        META_START,
        RE_META_ESCAPE,
        META_END,
        META_START,
        META_END,
        META_END,
    ),
    re.VERBOSE,
)

# A more human-readable version of the above RE is:
#
# re.compile(r"""(?:\\{|\\}|[^{}])+ # One or more of anything other than
#                                   # unescaped { or }
#                                   #
#                                 | # or
#                                   #
#              {(?:\\{|\\}|[^{}])*} # Anything of the form {X} where X
#                                   # doesn't contain unescaped { or }
#             """, re.VERBOSE)


WORD_RX = re.compile(r"(?:\d+(?:[.,]\d+)+|[\'\w]+[-\w\']*|[^\w\s]+)\s*", re.UNICODE)


class RetroFormatter:
    """Helper for iterating over the result of previous translations.

    Support iterating over previous actions, text, fragments of text, or words:

    text     : "Something something, blah! Blah: 45.8... (blah: foo42)   "
    fragments: "__________-----------______------________-------_________"
    words    : "__________---------__----__----__----____-____--_____----"

    """

    FRAGMENT_RX = re.compile(r"\s*[^\s]+\s*|^\s*$")

    def __init__(self, previous_translations):
        self.previous_translations = previous_translations

    def iter_last_actions(self):
        """Iterate over past actions (last first)."""
        for translation in reversed(self.previous_translations):
            yield from reversed(translation.formatting)

    def iter_last_fragments(self):
        """Iterate over last text fragments (last first).

        A text fragment is a series of non-whitespace characters
        followed by zero or more trailing whitespace characters.
        """
        replace = 0
        next_action = None
        current_fragment = ""
        for action in self.iter_last_actions():
            part = "" if action.text is None else action.text
            if (
                next_action is not None
                and next_action.text is not None
                and not next_action.prev_attach
            ):
                part += next_action.space_char
            if replace:
                # Ignore replaced content.
                if len(part) > replace:
                    part = part[:-replace]
                    replace = 0
                else:
                    replace -= len(part)
                    part = ""
            if part:
                # Find out new complete fragments.
                fragments = self.FRAGMENT_RX.findall(part + current_fragment)
                yield from reversed(fragments[1:])
                current_fragment = fragments[0]
            replace += len(action.prev_replace)
            next_action = action
        # Don't forget to process the current (first) fragment.
        if not current_fragment.isspace():
            yield current_fragment.lstrip()

    def last_fragments(self, count=1):
        """Return the last <count> text fragments."""
        fragment_list = []
        for fragment in self.iter_last_fragments():
            fragment_list.insert(0, fragment)
            if len(fragment_list) == count:
                break
        return fragment_list

    def iter_last_words(self, strip=False, rx=WORD_RX):
        """Iterate over last words (last first).

        If <strip> is False, then trailing whitespace is included
        as part of each word (useful for calculating position).

        For <strip> to be properly supported when a custom regexp is
        passed as <rx>, then it must include trailing whitespace as
        part of each word.
        """
        assert not rx.groups
        for fragment in self.iter_last_fragments():
            # Split each fragment into words.
            for word in reversed(rx.findall(fragment)):
                yield word.rstrip() if strip else word

    def last_words(self, count=1, strip=False, rx=WORD_RX):
        """Return the last <count> words."""
        word_list = []
        for w in self.iter_last_words(strip=strip, rx=rx):
            word_list.insert(0, w)
            if len(word_list) == count:
                break
        return word_list

    def last_text(self, size):
        """Return the last <size> characters."""
        text = ""
        if not size:
            return text
        for fragment in self.iter_last_fragments():
            text = fragment + text
            if len(text) >= size:
                break
        return text[-size:]


class _Context(RetroFormatter):
    """Context for formatting translations to actions.

    Keep tracks of previous actions as well as newly translated actions,
    offer helpers for creating new actions and convenient access to past
    actions/text/words.
    """

    def __init__(self, previous_translations, last_action):
        super().__init__(previous_translations)
        assert last_action is not None
        self.last_action = last_action
        self.translated_actions = []

    def new_action(self):
        """Create a new action, only copying global state."""
        return self.last_action.new_state()

    def copy_last_action(self):
        """Create a new action, cloning the last action state."""
        return self.last_action.copy_state()

    def translated(self, action):
        """Mark an action as translated."""
        assert action is not None
        self.translated_actions.append(action)
        self.last_action = action

    def iter_last_actions(self):
        """Custom iterator with support for newly translated actions."""
        yield from reversed(self.translated_actions)
        yield from super().iter_last_actions()


class Formatter:
    """Convert translations into output.

    The main entry point for this class is format, which takes in translations
    to format. Output is sent via an output class passed in through set_output.
    Other than setting the output, the formatter class is stateless.

    The output class can define the following functions, which will be called
    if available:

    send_backspaces -- Takes a number and deletes back that many characters.

    send_string -- Takes a string and prints it verbatim.

    send_key_combination -- Takes a string the dictionary format for specifying
    key combinations and issues them.

    send_engine_command -- Takes a string which names the special command to
    execute.

    """

    output_type = namedtuple(
        "output",
        [
            "send_backspaces",
            "send_string",
            "send_key_combination",
            "send_engine_command",
        ],
    )

    def __init__(self):
        self.set_output(None)
        self.spaces_after = False
        self.last_output_spaces_after = False
        self.start_capitalized = False
        self.start_attached = False
        self.space_char = " "
        self._listeners = set()

    def add_listener(self, callback):
        """Add a listener for translation outputs.

        Arguments:

        callback -- A function that takes: a list of translations to undo, a
        list of new translations to render, and a translation that is the
        context for the new translations.

        """
        self._listeners.add(callback)

    def remove_listener(self, callback):
        """Remove a listener added by add_listener."""
        self._listeners.remove(callback)

    def set_output(self, output):
        """Set the output class."""
        noop = lambda x: None
        output_type = self.output_type
        fields = output_type._fields
        self._output = output_type(*[getattr(output, f, noop) for f in fields])

    def set_space_placement(self, s):
        # Set whether spaces will be inserted
        # before the output or after the output
        self.spaces_after = bool(s == "After Output")

    def last_action(self, previous_translations):
        if previous_translations and previous_translations[-1].formatting:
            return previous_translations[-1].formatting[-1]
        return _Action(
            next_attach=self.start_attached or self.spaces_after,
            next_case=Case.CAP_FIRST_WORD if self.start_capitalized else None,
            space_char=self.space_char,
        )

    def format(self, undo, do, prev):
        """Format the given translations.

        Arguments:

        undo -- A sequence of translations that should be undone. The
        formatting parameter of the translations will be used to undo the
        actions that were taken, if possible.

        do -- The new actions to format. The formatting attribute will be
        filled in with the result.

        prev -- The last translation before the new actions in do. This
        translation's formatting attribute provides the context for the new
        rendered translations. If there is no context then this may be None.

        """
        assert undo or do

        if do:
            last_action = self.last_action(prev)
            ctx = _Context(prev or (), last_action)
            for t in do:
                if t.english:
                    t.formatting = _translation_to_actions(t.english, ctx)
                else:
                    t.formatting = _raw_to_actions(t.rtfcre[0], ctx)
            new = ctx.translated_actions
        else:
            new = []

        old = [a for t in undo for a in t.formatting]

        # Take into account previous look-ahead actions.

        if prev:
            text = ""
            for a in new:
                if a.text:
                    text = a.text
                    break
            tail = []
            for a in RetroFormatter(prev).iter_last_actions():
                if isinstance(a, _LookAheadAction):
                    old_a, new_a = a.action, a.update(text)
                    # Does the look-ahead action need updating?
                    if new_a == old_a:
                        # No, we're done.
                        break
                    old[0:0] = [old_a] + tail
                    new[0:0] = [new_a] + tail
                    text = a.text
                    tail = []
                    # Updating this action can impact another
                    # previous look-ahead action, keep going.
                elif a.text is not None:
                    if a.text:
                        # Stop when encountering a non-empty text action.
                        break
                    tail.insert(0, a)

        # Figure out what really changed.

        min_length = min(len(old), len(new))
        for i in range(min_length):
            if old[i] != new[i]:
                break
        else:
            i = min_length

        if i > 0:
            optimized_away = old[:i]
            old = old[i:]
            new = new[i:]
        else:
            optimized_away = []

        # Notify listeners.

        for callback in self._listeners:
            callback(old, new)

        # Render output.

        if optimized_away:
            last_action = optimized_away[-1]
        elif prev and prev[-1].formatting:
            last_action = prev[-1].formatting[-1]
        else:
            last_action = None

        OutputHelper(
            self._output, self.last_output_spaces_after, self.spaces_after
        ).render(last_action, old, new)
        self.last_output_spaces_after = self.spaces_after


class TextFormatter:
    """Format a series of action into text."""

    def __init__(self, spaces_after):
        self.spaces_after = spaces_after
        # Initial replaced text.
        self.replaced_text = ""
        # New appended text.
        self.appended_text = ""
        self.trailing_space = ""

    def _render_action(self, action):
        if self.spaces_after and self.trailing_space:
            assert self.appended_text.endswith(self.trailing_space)
            self.appended_text = self.appended_text[: -len(self.trailing_space)]
        if action.prev_replace:
            replaced = len(action.prev_replace)
            appended = len(self.appended_text)
            if replaced > appended:
                assert action.prev_replace.endswith(self.appended_text)
                replaced -= appended
                if replaced > len(self.replaced_text):
                    assert action.prev_replace.endswith(self.replaced_text)
                    self.replaced_text = action.prev_replace[:replaced]
                else:
                    assert self.replaced_text.endswith(action.prev_replace)
                    self.replaced_text = self.replaced_text[:-replaced]
                    self.replaced_text += action.prev_replace[:replaced]
                self.appended_text = ""
            else:
                assert self.appended_text.endswith(action.prev_replace)
                self.appended_text = self.appended_text[:-replaced]
        if not action.prev_attach:
            self.appended_text += action.space_char
        self.appended_text += action.text
        if self.spaces_after and not action.next_attach:
            self.appended_text += action.space_char
            self.trailing_space = action.space_char
        else:
            self.trailing_space = ""

    def render(self, action_list, last_action):
        """Render a series of action.

        Note: the function is a generator that yields non-text
        actions (commands, combos, ...) for special processing.
        """
        if self.spaces_after and last_action is not None:
            self.trailing_space = last_action.trailing_space
            self.appended_text = last_action.trailing_space
        for action in action_list:
            if action.text is None:
                yield action
            else:
                self._render_action(action)

    def reset(self, trailing_space):
        """Reset current state (rendered text)."""
        self.replaced_text = ""
        self.appended_text = trailing_space


class OutputHelper:
    """A helper class for minimizing the amount of change on output.

    This class figures out the current state, compares it to the new output and
    optimizes away extra backspaces and typing.

    """

    def __init__(self, output, before_spaces_after, after_spaces_after):
        self.output = output
        self.before = TextFormatter(before_spaces_after)
        self.after = TextFormatter(after_spaces_after)

    def flush(self):
        # FIXME:
        # - what about things like emoji zwj sequences?
        # - normalize strings to better handle combining characters?
        #
        # >>> u"C\u0327"
        # 'Ç'
        # >>> len(u"C\u0327")
        # 2
        # >>> len(unicodedata.normalize('NFC', u"C\u0327"))
        # 1
        if len(self.before.replaced_text) > len(self.after.replaced_text):
            assert self.before.replaced_text.endswith(self.after.replaced_text)
            replaced_text = self.before.replaced_text
        else:
            assert self.after.replaced_text.endswith(self.before.replaced_text)
            replaced_text = self.after.replaced_text
        before = (
            replaced_text[: len(replaced_text) - len(self.before.replaced_text)]
            + self.before.appended_text
        )
        after = (
            replaced_text[: len(replaced_text) - len(self.after.replaced_text)]
            + self.after.appended_text
        )
        common_length = len(commonprefix([before, after]))
        erased = len(before) - common_length
        if erased:
            self.output.send_backspaces(erased)
        appended = after[common_length:]
        if appended:
            self.output.send_string(appended)
        self.before.reset(self.after.trailing_space)
        self.after.reset(self.after.trailing_space)

    def render(self, last_action, undo, do):
        # Render undone actions, ignoring non-text actions.
        for action in self.before.render(undo, last_action):
            pass
        # Render new actions.
        for action in self.after.render(do, last_action):
            self.flush()
            if action.combo:
                self.output.send_key_combination(action.combo)
            elif action.command:
                self.output.send_engine_command(action.command)
        self.flush()


class _Action:
    """A hybrid class that stores instructions and resulting state.

    A single translation may be formatted into one or more actions. The
    instructions are used to render the current action and the state is used as
    context to render future translations.

    """

    def __init__(
        self,
        # Previous.
        prev_attach=False,
        prev_replace="",
        # Current.
        glue=False,
        word=None,
        orthography=True,
        space_char=" ",
        upper_carry=False,
        case=None,
        text=None,
        trailing_space="",
        word_is_finished=None,
        combo=None,
        command=None,
        # Next.
        next_attach=False,
        next_case=None,
    ):
        """Initialize a new action.

        Arguments:

        prev_attach -- True if there should be no space between this and the
                       previous action.

        prev_replace -- Text that should be deleted for this action.

        glue -- True if there be no space between this and the next action if
                the next action also has glue set to True.

        word -- The current root word (sans prefix, and un-cased). This is
                context for future actions whose behavior depends on it such as
                suffixes.

        upper_carry -- True if we are uppercasing the current word.

        word_is_finished -- True if word is finished.

        orthography -- True if orthography rules should be applies when adding
                      a suffix to this action.

        space_char -- this character will replace spaces after all other
        formatting has been applied

        case -- an integer to determine which case to output after formatting

        text -- The text that should be rendered for this action.

        trailing_space -- This the space that would be added when rendering
                          up to this action with space placement set to
                          'after output'.

        combo -- The key combo, in plover's key combo language, that should be
                 executed for this action.

        command -- The command that should be executed for this action.

        next_attach -- True if there should be no space between this and the next
                       action.

        next_case -- Case to apply to next action: capitalize/lower/upper...

        """
        # State variables
        self.prev_attach = prev_attach
        self.glue = glue
        self.word = word
        self.upper_carry = upper_carry
        self.orthography = orthography
        self.next_attach = next_attach
        self.next_case = next_case
        if word_is_finished is None:
            self.word_is_finished = not self.next_attach
        else:
            self.word_is_finished = word_is_finished
        # Persistent state variables
        self.space_char = space_char
        self.case = case
        self.trailing_space = trailing_space
        # Instruction variables
        self.prev_replace = prev_replace
        self.text = text
        self.combo = combo
        self.command = command

    def copy_state(self):
        """Clone this action but only clone the state variables."""
        return _Action(
            # Previous.
            prev_attach=self.next_attach,
            # Current.
            case=self.case,
            glue=self.glue,
            orthography=self.orthography,
            space_char=self.space_char,
            upper_carry=self.upper_carry,
            word=self.word,
            trailing_space=self.trailing_space,
            word_is_finished=self.word_is_finished,
            # Next.
            next_attach=self.next_attach,
            next_case=self.next_case,
        )

    def new_state(self):
        return _Action(
            # Previous.
            prev_attach=self.next_attach,
            # Current.
            space_char=self.space_char,
            case=self.case,
            trailing_space=self.trailing_space,
            # Next.
        )

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        kwargs = [
            "%s=%r" % (k, v)
            for k, v in self.__dict__.items()
            if v != self.DEFAULT.__dict__[k]
        ]
        return "Action(%s)" % ", ".join(sorted(kwargs))

    def __repr__(self):
        return str(self)


_Action.DEFAULT = _Action()


class _LookAheadAction(_Action):
    def __init__(self, pattern, action1, action2):
        self.pattern = pattern
        self.action1 = action1
        self.action2 = action2
        self.action = None
        self.update("")

    def update(self, text):
        if re.match(self.pattern, text) is None:
            self.action = self.action2
        else:
            self.action = self.action1
        return self.action

    def __getattr__(self, name):
        return getattr(self.action, name)

    def __str__(self):
        return "LookAheadAction(%s)" % str(self.__dict__)


def _translation_to_actions(translation, ctx):
    """Create actions for a translation.

    Arguments:

    translation -- A string with the translation to render.

    ctx -- The context in which this translation is formatted.

    Returns: A list of actions.

    """
    # Reduce the translation to atoms. An atom is an irreducible string that is
    # either entirely a single meta command or entirely text containing no meta
    # commands.
    if translation.isdigit():
        # If a translation is only digits then glue it to neighboring digits.
        atoms = [_glue_translation(translation)]
    else:
        atoms = filter(None, (x.strip(" ") for x in ATOM_RE.findall(translation)))
    action_list = []
    for atom in atoms:
        action = _atom_to_action(atom, ctx)
        action_list.append(action)
        ctx.translated(action)
    if not action_list:
        action = ctx.copy_last_action()
        action_list = [action]
        ctx.translated(action)
    return action_list


def _raw_to_actions(stroke, ctx):
    """Turn a raw stroke into actions.

    Arguments:

    stroke -- A string representation of the stroke.

    ctx -- The context in which the new actions are created.

    Returns: A list of actions.

    """
    # If a raw stroke is composed of digits then remove the dash (if
    # present) and glue it to any neighboring digits. Otherwise, just
    # output the raw stroke as is.
    no_dash = stroke.replace("-", "", 1)
    if no_dash.isdigit():
        return _translation_to_actions(no_dash, ctx)
    action = _Action(
        text=stroke,
        word=stroke,
        case=ctx.last_action.case,
        prev_attach=ctx.last_action.next_attach,
        space_char=ctx.last_action.space_char,
        trailing_space=ctx.last_action.space_char,
    )
    ctx.translated(action)
    return [action]


def _meta_to_action(meta, ctx):
    meta_name, meta_arg = _parse_meta(meta)
    if meta_name is not None:
        meta_fn = registry.get_plugin("meta", meta_name).obj
        action = meta_fn(ctx, meta_arg)
    else:
        action = ctx.new_action()
    return action


def _atom_to_action(atom, ctx):
    """Convert an atom into an action.

    Arguments:

    atom -- A string holding an atom. An atom is an irreducible string that is
    either entirely a single meta command or entirely text containing no meta
    commands.

    ctx -- The context in which the new action takes place.

    Returns: An action for the atom.

    """
    meta = _get_meta(atom)
    if meta is not None:
        meta = _unescape_atom(meta)
        action = _meta_to_action(meta, ctx)
    else:
        action = ctx.new_action()
        action.text = _unescape_atom(atom)
    _finalize_action(action, ctx)
    return action


def _finalize_action(action, ctx):
    """Finalize action's text."""
    if isinstance(action, _LookAheadAction):
        _finalize_action(action.action1, ctx)
        _finalize_action(action.action2, ctx)
        return
    text = action.text
    if text is None:
        return
    # Update word.
    if action.word is None:
        last_word = None
        if action.glue and ctx.last_action.glue:
            last_word = ctx.last_action.word
        action.word = rightmost_word((last_word or "") + text)
    # Apply case.
    case = ctx.last_action.next_case
    if case is None and action.prev_attach and ctx.last_action.upper_carry:
        case = Case.UPPER_FIRST_WORD
    text = apply_case(text, case)
    if case == Case.UPPER_FIRST_WORD:
        action.upper_carry = not has_word_boundary(text)
    # Apply mode.
    action.text = apply_mode(
        text, action.case, action.space_char, action.prev_attach, ctx.last_action
    )
    # Update trailing space.
    action.trailing_space = "" if action.next_attach else action.space_char


def apply_case(text, case):
    if case is None:
        return text
    if case == Case.CAP_FIRST_WORD:
        return capitalize_first_word(text)
    if case == Case.LOWER_FIRST_CHAR:
        return lower_first_character(text)
    if case == Case.UPPER_FIRST_WORD:
        return upper_first_word(text)
    raise ValueError("%r is not a valid case" % case)


def apply_mode(text, case, space_char, begin, last_action):
    # Should title case be applied to the beginning of the next string?
    lower_title_case = begin and not last_action.case in (
        Case.CAP_FIRST_WORD,
        Case.UPPER_FIRST_WORD,
    )
    # Apply case, then replace space character.
    text = apply_mode_case(text, case, lower_title_case)
    text = apply_mode_space_char(text, space_char)
    # Title case is sensitive to lower flag.
    if last_action.next_case == Case.LOWER_FIRST_CHAR and text and case == Case.TITLE:
        text = lower_first_character(text)
    return text


def apply_mode_case(text, case, appended):
    if case is None:
        return text
    if case == Case.LOWER:
        return text.lower()
    if case == Case.UPPER:
        return text.upper()
    if case == Case.TITLE:
        # Do nothing to appended output
        if appended:
            return text
        return capitalize_all_words(text)
    raise ValueError("%r is not a valid case" % case)


def apply_mode_space_char(text, space_char):
    if space_char == SPACE:
        return text
    return text.replace(SPACE, space_char)


def _get_meta(atom):
    """Return the meta command, if any, without surrounding meta markups."""
    if atom is not None and atom.startswith(META_START) and atom.endswith(META_END):
        return atom[len(META_START) : -len(META_END)]
    return None


def _glue_translation(s):
    """Mark the given string as a glue stroke."""
    return META_START + META_GLUE_FLAG + s + META_END


def _unescape_atom(atom):
    """Replace escaped meta markups with unescaped meta markups."""
    atom = atom.replace(META_ESC_START, META_START)
    atom = atom.replace(META_ESC_END, META_END)
    return atom


def capitalize_first_word(s):
    """Capitalize the first letter of s.

    - 'foo bar' -> 'Foo bar'
    - 'STUFF' -> 'STUFF'
    """
    return s[0:1].upper() + s[1:]


def capitalize_all_words(s):
    """Capitalize each word of s.

    - 'foo bar' -> 'Foo Bar'
    - "O'Something STUFF" -> "O'something Stuff"
    """
    return string.capwords(s, SPACE)


def lower_first_character(s):
    """Lowercase the first letter of s."""
    return s[0:1].lower() + s[1:]


def upper_all_words(s):
    """Uppercase the entire s."""
    return s.upper()


def upper_first_word(s):
    """Uppercase first word of s."""
    m = WORD_RX.match(s)
    if m is None:
        return s
    first_word = m.group()
    return first_word.upper() + s[len(first_word) :]


def rightmost_word(s):
    """Get the rightmost word in s."""
    words = WORD_RX.findall(s)
    if not words:
        return ""
    last_word = words[-1]
    if last_word[-1].isspace():
        return ""
    return last_word


def has_word_boundary(s):
    """Return True if s contains a word boundary
    (e.g.: more than 1 word, or white space).
    """
    if not s:
        return False
    if s[0].isspace() or s[-1].isspace():
        return True
    return len(WORD_RX.findall(s)) > 1



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Open Steno Project
# See LICENSE.txt for details.

"""Tests for misc.py."""

import inspect
import os

import pytest

import plover.misc as misc
import plover.oslayer.config as conf
from plover.resource import ASSET_SCHEME

from plover_build_utils.testing import parametrize


def test_popcount_8():
    for n in range(256):
        assert misc.popcount_8(n) == format(n, "b").count("1")
    with pytest.raises(AssertionError):
        misc.popcount_8(256)
    with pytest.raises(AssertionError):
        misc.popcount_8(-1)


if conf.PLATFORM == "win":
    ABS_PATH = os.path.normcase(r"c:\foo\bar")
else:
    ABS_PATH = "/foo/bar"


@parametrize(
    (
        # Asset, no change.
        lambda: (ASSET_SCHEME + "foo:bar", ASSET_SCHEME + "foo:bar"),
        # Absolute path, no change.
        lambda: (ABS_PATH, ABS_PATH),
        # Relative path, resolve relative to configuration directory.
        lambda: (
            os.path.normcase(os.path.normpath("foo/bar")),
            os.path.normcase(
                os.path.join(os.path.realpath(conf.CONFIG_DIR), "foo", "bar")
            ),
        ),
        # Path below the user home directory.
        lambda: (
            os.path.normcase(os.path.normpath("~/foo/bar")),
            os.path.normcase(os.path.expanduser(os.path.normpath("~/foo/bar"))),
        ),
    )
)
def test_dictionary_path(short_path, full_path):
    for input, function, expected in (
        # Unchanged.
        (short_path, "shorten", short_path),
        (full_path, "expand", full_path),
        # Shorten.
        (full_path, "shorten", short_path),
        # Expand.
        (short_path, "expand", full_path),
    ):
        function = "%s_path" % function
        result = getattr(misc, function)(input)
        assert result == expected, function


@parametrize(
    (
        # Boolean.
        lambda: (False, False),
        lambda: (True, True),
        # True string values.
        lambda: ("1", True),
        lambda: ("yes", True),
        lambda: ("true", True),
        lambda: ("on", True),
        # False string values.
        lambda: ("0", False),
        lambda: ("no", False),
        lambda: ("false", False),
        lambda: ("off", False),
        # Invalid string values.
        lambda: ("yep", ValueError),
        lambda: ("nope", ValueError),
        # Other types.
        lambda: (0, False),
        lambda: (1, True),
        lambda: (42, True),
        lambda: (4.2, True),
        lambda: (0.0, False),
        lambda: (None, False),
    )
)
def test_boolean(input, output):
    if inspect.isclass(output):
        with pytest.raises(output):
            misc.boolean(input)
    else:
        assert misc.boolean(input) == output


def test_to_surrogate_pairs():
    # Split unicode characters above 0xFFFF
    assert misc.to_surrogate_pair(chr(0x1F4A6)) == [0xD83D, 0xDCA6]
    # Do not slit characters below 0xFFFF
    assert misc.to_surrogate_pair(chr(0x20)) == [0x20]
    # Do not split already split characters.
    assert misc.to_surrogate_pair(chr(0xD83D) + chr(0xDCA6)) == [0xD83D, 0xDCA6]



#!/usr/bin/env python3

import fnmatch
import os
import py_compile
import shutil
import sys


def source_less(directory, excludes=()):
    for dirpath, dirnames, filenames in os.walk(directory):
        if "__pycache__" in dirnames:
            dirnames.remove("__pycache__")
            cache = os.path.join(dirpath, "__pycache__")
            shutil.rmtree(cache)
        for name in filenames:
            if not name.endswith(".py"):
                continue
            py = os.path.join(dirpath, name)
            for pattern in excludes:
                if fnmatch.fnmatch(py, pattern):
                    break
            else:
                pyc = py + "c"
                py_compile.compile(py, cfile=pyc)
                os.unlink(py)


if __name__ == "__main__":
    directory = sys.argv[1]
    excludes = sys.argv[2:]
    source_less(directory, excludes)



#!/usr/bin/env python3

from urllib.request import urlopen
from urllib.parse import urlsplit
import hashlib
import os
import sys


DOWNLOADS_DIR = os.path.join(".cache", "downloads")


def download(url, sha1=None, filename=None, downloads_dir=DOWNLOADS_DIR):
    if filename is None:
        filename = os.path.basename(urlsplit(url).path)
    dst = os.path.join(downloads_dir, filename)
    dst_dir = os.path.dirname(dst)
    if not os.path.exists(dst_dir):
        os.makedirs(dst_dir)
    retries = 0
    while retries < 2:
        if sha1 is None or not os.path.exists(dst):
            retries += 1
            try:
                with urlopen(url) as req, open(dst, "wb") as fp:
                    fp.write(req.read())
            except Exception as e:
                print("error", e, file=sys.stderr)
                continue
        if sha1 is None:
            break
        h = hashlib.sha1()
        with open(dst, "rb") as fp:
            while True:
                d = fp.read(4 * 1024 * 1024)
                if not d:
                    break
                h.update(d)
        if h.hexdigest() == sha1:
            break
        print(
            "sha1 does not match: %s instead of %s" % (h.hexdigest(), sha1),
            file=sys.stderr,
        )
        os.unlink(dst)
    assert os.path.exists(dst), "could not successfully retrieve %s" % url
    return dst


if __name__ == "__main__":
    args = sys.argv[1:]
    url = args.pop(0)
    sha1 = None
    filename = None
    downloads_dir = DOWNLOADS_DIR
    if args:
        sha1 = args.pop(0) or None
    if args:
        filename = args.pop(0) or None
    if args:
        downloads_dir = args.pop(0) or DOWNLOADS_DIR
    print(download(url, sha1, filename, downloads_dir))



from contextlib import contextmanager
from pathlib import Path
import os
import tempfile


@contextmanager
def make_dict(tmp_path, contents, extension=None, name=None):
    kwargs = {"dir": str(tmp_path)}
    if name is not None:
        kwargs["prefix"] = name + "_"
    if extension is not None:
        kwargs["suffix"] = "." + extension
    fd, path = tempfile.mkstemp(**kwargs)
    try:
        os.write(fd, contents)
        os.close(fd)
        yield Path(path)
    finally:
        os.unlink(path)



import os

import pytest

# Ensure i18n support does not mess us up.
os.environ["LANGUAGE"] = "C"

from plover import system
from plover.config import DEFAULT_SYSTEM_NAME
from plover.registry import registry


@pytest.fixture(scope="session", autouse=True)
def setup_plover():
    registry.update()
    system.setup(DEFAULT_SYSTEM_NAME)


pytest.register_assert_rewrite("plover_build_utils.testing")


def pytest_collection_modifyitems(items):
    for item in items:
        # Mark `gui_qt` tests.
        if "gui_qt" in item.location[0].split(os.path.sep):
            item.add_marker(pytest.mark.gui_qt)



#!/usr/bin/env python3
# Copyright (c) 2010 Joshua Harlan Lifton.
# See LICENSE.txt for details.

import os
import re
import subprocess
import sys

from setuptools import setup
from packaging.version import Version

sys.path.insert(0, os.path.dirname(__file__))

__software_name__ = "plover"

with open(os.path.join(__software_name__, "__init__.py")) as fp:
    exec(fp.read())

from plover_build_utils.setup import BuildPy, BuildUi, Command, Develop, babel_options


Develop.build_dependencies.append("build_py")
BuildPy.build_dependencies.append("build_ui")
cmdclass = {
    "build_py": BuildPy,
    "build_ui": BuildUi,
    "develop": Develop,
}
options = {}

PACKAGE = "%s-%s" % (
    __software_name__,
    __version__,
)

# Helpers. {{{


def get_version():
    if not os.path.exists(".git"):
        return None

    version = __version__

    # extend version with git revision if no tag is available - used for builds during development
    git_version = (
        subprocess.check_output("git describe --tags --match=v[0-9]*".split())
        .strip()
        .decode()
    )
    m = re.match(r"^v(\d[\d.]*(?:(?:\.dev|rc)\d+)?)(-\d+-g[a-f0-9]*)?$", git_version)
    assert m is not None, git_version
    if m.group(2) is not None:
        version += "+" + m.group(2)[1:].replace("-", ".")

    return version


# }}}

# `bdist_win` command. {{{


class BinaryDistWin(Command):
    description = "create distribution(s) for MS Windows"
    user_options = [
        ("trim", "t", "trim the resulting distribution to reduce size"),
        ("zipdir", "z", "create a zip of the resulting directory"),
        (
            "installer",
            "i",
            "create an executable installer for the resulting distribution",
        ),
        ("bash=", None, "bash executable to use for running the build script"),
    ]
    boolean_options = ["installer", "trim", "zipdir"]
    extra_args = []

    def initialize_options(self):
        self.bash = None
        self.installer = False
        self.trim = False
        self.zipdir = False

    def finalize_options(self):
        pass

    def run(self):
        cmd = [self.bash or "bash", "windows/dist_build.sh"]
        if self.installer:
            cmd.append("--installer")
        if self.trim:
            cmd.append("--trim")
        if self.zipdir:
            cmd.append("--zipdir")
        cmd.extend((__software_name__, __version__, self.bdist_wheel()))
        if self.verbose:
            print("running", " ".join(cmd))
        subprocess.check_call(cmd)


if sys.platform.startswith("win32"):
    cmdclass["bdist_win"] = BinaryDistWin

# }}}

# `launch` command. {{{


class Launch(Command):
    description = "run %s from source" % __software_name__.capitalize()
    command_consumes_arguments = True
    user_options = []

    def initialize_options(self):
        self.args = None

    def finalize_options(self):
        pass

    def run(self):
        with self.project_on_sys_path():
            python_path = os.environ.get("PYTHONPATH", "").split(os.pathsep)
            python_path.insert(0, sys.path[0])
            os.environ["PYTHONPATH"] = os.pathsep.join(python_path)
            cmd = [sys.executable, "-m", "plover.scripts.main"] + self.args
            if sys.platform.startswith("win32"):
                # Workaround https://bugs.python.org/issue19066
                subprocess.Popen(cmd, cwd=os.getcwd())
                sys.exit(0)
            os.execv(cmd[0], cmd)


cmdclass["launch"] = Launch

# }}}

# `patch_version` command. {{{


class PatchVersion(Command):
    description = "patch package version from VCS"
    command_consumes_arguments = True
    user_options = []

    def initialize_options(self):
        self.args = []

    def finalize_options(self):
        assert 0 <= len(self.args) <= 1

    def patch_version(
        self, version_file_path, regex_pattern, regex_replacement, version
    ):
        with open(version_file_path, "r") as fp:
            contents = fp.read().split("\n")
        contents = [
            re.sub(regex_pattern, regex_replacement % version, line)
            for line in contents
        ]
        with open(version_file_path, "w") as fp:
            fp.write("\n".join(contents))

    def run(self):
        if self.args:
            version = self.args[0]
            # Ensure it's valid.
            Version(version)
        else:
            version = get_version()
            if version is None:
                sys.exit(1)
        if self.verbose:
            print("patching version to", version)

        plover_init_file_path = os.path.join("plover", "__init__.py")
        self.patch_version(
            plover_init_file_path, r"^__version__ = .*$", '__version__ = "%s"', version
        )

        doc_conf_file_path = os.path.join("doc", "conf.py")
        self.patch_version(
            doc_conf_file_path, r"^release = .*$", 'release = "%s"', version
        )


cmdclass["patch_version"] = PatchVersion

# }}}

# `bdist_app` and `bdist_dmg` commands. {{{


class BinaryDistApp(Command):
    description = "create an application bundle for Mac"
    user_options = []
    extra_args = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        cmd = ["bash", "osx/make_app.sh", self.bdist_wheel()]
        if self.verbose:
            print("running", " ".join(cmd))
        subprocess.check_call(cmd)


class BinaryDistDmg(Command):
    user_options = [
        (
            "skip-app-build",
            None,
            "skip building the app; assume dist/Plover.app already exists",
        ),
    ]
    boolean_options = ["skip-app-build"]

    def initialize_options(self):
        self.skip_app_build = False

    def finalize_options(self):
        pass

    def run(self):
        # Build the app first unless the user asked for DMG-only.
        if not self.skip_app_build:
            self.run_command("bdist_app")

        app_path = "dist/Plover.app"
        if not os.path.isdir(app_path):
            raise SystemExit(
                "Plover.app not found in 'dist/'. Build the app first or omit --skip-app-build."
            )

        # Determine targeted macOS platform
        from setuptools.command.bdist_wheel import get_platform

        platform = get_platform(app_path)

        cmd = ["bash", "osx/make_dmg.sh", platform, PACKAGE]
        if self.verbose:
            print("running", " ".join(cmd))
        subprocess.check_call(cmd)


class NotarizeApp(Command):
    user_options = []
    extra_args = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        cmd = ["bash", "osx/notarize_app.sh"]
        if self.verbose:
            print("running", " ".join(cmd))
        subprocess.check_call(cmd)


class NotarizeDmg(Command):
    user_options = []
    extra_args = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import glob

        # Locate the DMG we want to notarize
        dmgs = sorted(glob.glob("dist/*.dmg"))
        if not dmgs:
            raise SystemExit(
                "No DMG found in 'dist/'. Build the DMG first (bdist_dmg)."
            )
        if len(dmgs) > 1:
            raise SystemExit(
                "Multiple DMGs found in 'dist/': "
                + ", ".join(os.path.basename(p) for p in dmgs)
            )
        dmg_path = dmgs[0]

        cmd = ["bash", "osx/notarize_dmg.sh", dmg_path]
        if self.verbose:
            print("running", " ".join(cmd))
        subprocess.check_call(cmd)


if sys.platform.startswith("darwin"):
    cmdclass["bdist_app"] = BinaryDistApp
    cmdclass["bdist_dmg"] = BinaryDistDmg
    cmdclass["notarize_app"] = NotarizeApp
    cmdclass["notarize_dmg"] = NotarizeDmg

# }}}

# `bdist_appimage` command. {{{


class BinaryDistAppImage(Command):
    description = "create AppImage distribution for Linux"
    user_options = [
        ("docker", None, "use docker to run the build script"),
        (
            "no-update-tools",
            None,
            "don't try to update AppImage tools, only fetch missing ones",
        ),
    ]
    boolean_options = ["docker", "no-update-tools"]

    def initialize_options(self):
        self.docker = False
        self.no_update_tools = False

    def finalize_options(self):
        pass

    def run(self):
        cmd = ["./linux/appimage/build.sh"]
        if self.docker:
            cmd.append("--docker")
        else:
            cmd.extend(("--python", sys.executable))
        if self.no_update_tools:
            cmd.append("--no-update-tools")
        cmd.extend(("--wheel", self.bdist_wheel()))
        if self.verbose:
            print("running", " ".join(cmd))
        subprocess.check_call(cmd)


if sys.platform.startswith("linux"):
    cmdclass["bdist_appimage"] = BinaryDistAppImage

# }}}

# i18n support. {{{

options.update(babel_options(__software_name__))

BuildPy.build_dependencies.append("compile_catalog")
BuildUi.hooks.append("plover_build_utils.pyqt:gettext")

# }}}


def reqs(name):
    with open(os.path.join("reqs", name + ".txt")) as fp:
        return fp.read()


setup(
    name=__software_name__,
    version=__version__,
    description=__description__,
    url=__url__,
    download_url=__download_url__,
    license=__license__,
    options=options,
    cmdclass=cmdclass,
    install_requires=reqs("dist"),
    extras_require={
        "gui_qt": reqs("dist_extra_gui_qt"),
        "log": reqs("dist_extra_log"),
    },
)

# vim: foldmethod=marker



import collections

from plover.steno import sort_steno_strokes


Suggestion = collections.namedtuple("Suggestion", "text steno_list")


class Suggestions:
    def __init__(self, dictionary):
        self.dictionary = dictionary

    def find(self, translation):
        suggestions = []

        mods = [
            "%s",  # Same
            "{^%s}",  # Prefix
            "{^}%s",
            "{^%s^}",  # Infix
            "{^}%s{^}",
            "{%s^}",  # Suffix
            "%s{^}",
            "{&%s}",  # Fingerspell
            "{#%s}",  # Command
        ]

        possible_translations = {translation}

        # Only strip spaces, so patterns with \n or \t are correctly handled.
        stripped_translation = translation.strip(" ")
        if stripped_translation and stripped_translation != translation:
            possible_translations.add(stripped_translation)

        lowercase_translation = translation.lower()
        if lowercase_translation != translation:
            possible_translations.add(lowercase_translation)

        similar_words = self.dictionary.casereverse_lookup(translation.lower())
        if similar_words:
            possible_translations |= set(similar_words)

        for t in possible_translations:
            for modded_translation in [mod % t for mod in mods]:
                strokes_list = self.dictionary.reverse_lookup(modded_translation)
                if not strokes_list:
                    continue
                strokes_list = sort_steno_strokes(strokes_list)
                suggestion = Suggestion(modded_translation, strokes_list)
                suggestions.append(suggestion)

        return suggestions



# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

"""A module to handle logging."""

import os
import sys
import logging
import traceback

from logging.handlers import RotatingFileHandler
from logging import DEBUG, INFO, WARNING, ERROR

from plover.oslayer.config import CONFIG_DIR, PLATFORM


LOG_FORMAT = "%(asctime)s [%(threadName)s] %(levelname)s: %(message)s"
LOG_FILENAME = os.path.realpath(os.path.join(CONFIG_DIR, "plover.log"))
LOG_MAX_BYTES = 10000000
LOG_COUNT = 9

STROKE_LOG_FORMAT = "%(asctime)s %(message)s"


class NoExceptionTracebackFormatter(logging.Formatter):
    """Custom formatter for formatting exceptions without traceback."""

    def format(self, record):
        # Calls to formatException are cached.
        # (see http://bugs.python.org/issue1295)
        orig_exc_text = record.exc_text
        record.exc_text = None
        try:
            return super().format(record)
        finally:
            record.exc_text = orig_exc_text

    def formatException(self, exc_info):
        etype, evalue, tb = exc_info
        lines = traceback.format_exception_only(etype, evalue)
        return "".join(lines)


class FileHandler(RotatingFileHandler):
    def __init__(self, filename=LOG_FILENAME, format=LOG_FORMAT):
        super().__init__(
            filename, maxBytes=LOG_MAX_BYTES, backupCount=LOG_COUNT, encoding="utf-8"
        )
        self.setFormatter(logging.Formatter(format))


class PrintHandler(logging.StreamHandler):
    """Handler using L{print_} to output messages."""

    def __init__(self, format=LOG_FORMAT):
        super().__init__(sys.stderr)
        self.setFormatter(logging.Formatter(format))


class Logger:
    def __init__(self):
        self._print_handler = PrintHandler()
        self._print_handler.setLevel(WARNING)
        self._file_handler = None
        self._platform_handler = None
        self._logger = logging.getLogger("plover")
        self._logger.addHandler(self._print_handler)
        self._logger.setLevel(INFO)
        self._stroke_filename = None
        self._stroke_logger = logging.getLogger("plover-strokes")
        self._stroke_logger.setLevel(INFO)
        self._stroke_handler = None
        self._log_strokes = False
        self._log_translations = False

    def has_platform_handler(self):
        return self._platform_handler is not None

    def setup_platform_handler(self):
        if self.has_platform_handler():
            return
        NotificationHandler = None
        try:
            from plover.oslayer.log import NotificationHandler
        except Exception:
            self.info("could not import platform gui log", exc_info=True)
            return
        try:
            handler = NotificationHandler()
            self.addHandler(handler)
        except Exception:
            self.info("could not initialize platform gui log", exc_info=True)
            return
        self._platform_handler = handler

    def set_level(self, level):
        self._print_handler.setLevel(level)
        if self._file_handler is not None:
            self._file_handler.setLevel(level)
        self.setLevel(level)

    def setup_logfile(self):
        assert self._file_handler is None
        self._file_handler = FileHandler()
        self._file_handler.setLevel(self.level)
        self._logger.addHandler(self._file_handler)

    def _setup_stroke_logging(self):
        is_logging = self._stroke_handler is not None
        must_log = (
            self._log_strokes or self._log_translations
        ) and self._stroke_filename is not None
        if must_log:
            if is_logging:
                start_logging = stop_logging = (
                    self._stroke_filename != self._stroke_handler.baseFilename
                )
            else:
                stop_logging = False
                start_logging = True
        else:
            stop_logging = is_logging
            start_logging = False
        if stop_logging:
            self._stroke_logger.removeHandler(self._stroke_handler)
            self._stroke_handler.close()
            self._stroke_handler = None
        if start_logging:
            self._stroke_handler = FileHandler(
                filename=self._stroke_filename, format=STROKE_LOG_FORMAT
            )
            self._stroke_logger.addHandler(self._stroke_handler)

    def set_stroke_filename(self, filename=None):
        if filename is not None:
            filename = os.path.realpath(filename)
        if self._stroke_filename == filename:
            return
        assert filename != LOG_FILENAME
        self.info("set_stroke_filename(%s)", filename)
        self._stroke_filename = filename
        self._setup_stroke_logging()

    def enable_stroke_logging(self, enable):
        if self._log_strokes == enable:
            return
        self.info("enable_stroke_logging(%s)", enable)
        self._log_strokes = enable
        self._setup_stroke_logging()

    def enable_translation_logging(self, enable):
        if self._log_translations == enable:
            return
        self.info("enable_translation_logging(%s)", enable)
        self._log_translations = enable
        self._setup_stroke_logging()

    def log_stroke(self, stroke):
        if not self._log_strokes or self._stroke_handler is None:
            return
        self._stroke_logger.info("%s", stroke)

    def log_translation(self, undo, do, prev):
        if not self._log_translations or self._stroke_handler is None:
            return
        # TODO: Figure out what to actually log here.
        for u in undo:
            self._stroke_logger.info("*%s", u)
        for d in do:
            self._stroke_logger.info(d)

    # Delegate calls to _logger.
    def __getattr__(self, name):
        return getattr(self._logger, name)


# Set up default logger.
__logger = Logger()

# The following functions direct all input to __logger.
debug = __logger.debug
info = __logger.info
warning = __logger.warning
error = __logger.error
set_level = __logger.set_level
add_handler = __logger.addHandler
remove_handler = __logger.removeHandler
has_platform_handler = __logger.has_platform_handler
setup_platform_handler = __logger.setup_platform_handler
# Strokes/translation logging.
set_stroke_filename = __logger.set_stroke_filename
stroke = __logger.log_stroke
translation = __logger.log_translation
enable_stroke_logging = __logger.enable_stroke_logging
enable_translation_logging = __logger.enable_translation_logging
# Logfile support.
setup_logfile = __logger.setup_logfile



#!/usr/bin/env python3

import os
import sys
import zipfile


def zipdir(directory, compression=zipfile.ZIP_DEFLATED):
    zipname = "%s.zip" % directory
    prefix = os.path.dirname(directory)
    with zipfile.ZipFile(zipname, "w", compression) as zf:
        for dirpath, dirnames, filenames in os.walk(directory):
            for name in filenames:
                src = os.path.join(dirpath, name)
                dst = os.path.relpath(src, prefix)
                zf.write(src, dst)


if __name__ == "__main__":
    directory = sys.argv[1]
    zipdir(directory)



from collections import namedtuple, OrderedDict
from functools import wraps
from queue import Queue
import functools
import os
import shutil
import threading

from plover import log, system
from plover.dictionary.loading_manager import DictionaryLoadingManager
from plover.formatting import Formatter
from plover.misc import shorten_path
from plover.registry import registry
from plover.resource import ASSET_SCHEME, resource_filename
from plover.steno import Stroke
from plover.steno_dictionary import StenoDictionary, StenoDictionaryCollection
from plover.suggestions import Suggestions
from plover.translation import Translator


StartingStrokeState = namedtuple(
    "StartingStrokeState", "attach capitalize space_char", defaults=(False, False, " ")
)


MachineParams = namedtuple("MachineParams", "type options keymap")


class ErroredDictionary(StenoDictionary):
    """Placeholder for dictionaries that failed to load."""

    def __init__(self, path, exception):
        super().__init__()
        self.enabled = False
        self.readonly = True
        self.path = path
        self.exception = exception

    def __eq__(self, other):
        if not isinstance(other, ErroredDictionary):
            return False
        return (self.path, self.exception) == (other.path, other.exception)


def copy_default_dictionaries(dictionaries_files):
    """Recreate default dictionaries.

    Each default dictionary is recreated if it's
    in use by the current config and missing.
    """

    for dictionary in dictionaries_files:
        # Ignore assets.
        if dictionary.startswith(ASSET_SCHEME):
            continue
        # Nothing to do if dictionary file already exists.
        if os.path.exists(dictionary):
            continue
        # Check it's actually a default dictionary.
        basename = os.path.basename(dictionary)
        if basename not in system.DEFAULT_DICTIONARIES:
            continue
        default_dictionary = os.path.join(system.DICTIONARIES_ROOT, basename)
        log.info("recreating %s from %s", dictionary, default_dictionary)
        shutil.copyfile(resource_filename(default_dictionary), dictionary)


def with_lock(func):
    # To keep __doc__/__name__ attributes of the initial function.
    @wraps(func)
    def _with_lock(self, *args, **kwargs):
        with self:
            return func(self, *args, **kwargs)

    return _with_lock


class StenoEngine:
    HOOKS = """
    stroked
    translated
    machine_state_changed
    output_changed
    config_changed
    dictionaries_loaded
    dictionary_state_changed
    send_string
    send_backspaces
    send_key_combination
    add_translation
    focus
    configure
    lookup
    suggestions
    quit
    """.split()

    def __init__(self, config, controller, keyboard_emulation):
        self._config = config
        self._controller = controller
        self._is_running = False
        self._queue = Queue()
        self._lock = threading.RLock()
        self._machine = None
        self._machine_state = None
        self._machine_params = MachineParams(None, None, None)
        self._formatter = Formatter()
        self._formatter.set_output(
            Formatter.output_type(
                self._send_backspaces,
                self._send_string,
                self._send_key_combination,
                self._send_engine_command,
            )
        )
        self._formatter.add_listener(self._on_translated)
        self._translator = Translator()
        self._translator.add_listener(log.translation)
        self._translator.add_listener(self._formatter.format)
        self._dictionaries = self._translator.get_dictionary()  # type: StenoDictionaryCollection
        self._dictionaries_manager = DictionaryLoadingManager(
            functools.partial(self._trigger_hook, "dictionary_state_changed")
        )
        self._running_state = self._translator.get_state()
        self._translator.clear_state()
        self._keyboard_emulation = keyboard_emulation
        self._hooks = {hook: [] for hook in self.HOOKS}
        self._running_extensions = {}

    def __enter__(self):
        self._lock.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._lock.__exit__(exc_type, exc_value, traceback)

    def _in_engine_thread(self):
        raise NotImplementedError()

    def _same_thread_hook(self, func, *args, **kwargs):
        if self._in_engine_thread():
            func(*args, **kwargs)
        else:
            self._queue.put((func, args, kwargs))

    def run(self):
        while True:
            func, args, kwargs = self._queue.get()
            try:
                with self._lock:
                    if func(*args, **kwargs):
                        break
            except Exception:
                log.error("engine %s failed", func.__name__[1:], exc_info=True)

    def _on_control_message(self, msg):
        if msg[0] == "command":
            self._same_thread_hook(self._execute_engine_command, *msg[1:], force=True)
        else:
            log.error("ignoring invalid control message: %r", msg)

    def _stop(self):
        self._controller.stop()
        self._stop_extensions(self._running_extensions.keys())
        if self._machine is not None:
            self._machine.stop_capture()
            self._machine = None

    def _start(self):
        self._set_output(self._config["auto_start"])
        self._update(full=True)
        self._controller.start(self._on_control_message)

    def _set_dictionaries(self, dictionaries):
        def dictionaries_changed(l1, l2):
            if len(l1) != len(l2):
                return True
            for d1, d2 in zip(l1, l2):
                if d1 is not d2:
                    return True
            return False

        if not dictionaries_changed(dictionaries, self._dictionaries.dicts):
            # No change.
            return
        self._dictionaries.set_dicts(dictionaries)
        self._trigger_hook(
            "dictionaries_loaded", StenoDictionaryCollection(dictionaries)
        )

    def _update(self, config_update=None, full=False, reset_machine=False):
        original_config = self._config.as_dict()
        # Update configuration.
        if config_update is not None:
            self._config.update(**config_update)
            config = self._config.as_dict()
        else:
            config = original_config
        # Create configuration update.
        if full:
            config_update = config
        else:
            config_update = {
                option: value
                for option, value in config.items()
                if value != original_config[option]
            }
            # Save config if anything changed.
            if config_update:
                self._config.save()
        # Update logging.
        log.set_stroke_filename(config["log_file_name"])
        log.enable_stroke_logging(config["enable_stroke_logging"])
        log.enable_translation_logging(config["enable_translation_logging"])
        # Update output.
        self._formatter.set_space_placement(config["space_placement"])
        self._formatter.start_attached = config["start_attached"]
        self._formatter.start_capitalized = config["start_capitalized"]
        self._translator.set_min_undo_length(config["undo_levels"])
        self._keyboard_emulation.set_key_press_delay(config["time_between_key_presses"])
        # This only applies to UInput, because it emulates a physical keyboard and follows the layout set in software. Because there is no standard of defining it, the user has to do so manually if not using a QWERTY keyboard.
        if hasattr(self._keyboard_emulation, "_update_layout"):
            self._keyboard_emulation._update_layout(config["keyboard_layout"])
        # Update system.
        system_name = config["system_name"]
        if system.NAME != system_name:
            log.info("loading system: %s", system_name)
            system.setup(system_name)
        # Update machine.
        update_keymap = False
        start_machine = False
        machine_params = MachineParams(
            config["machine_type"],
            config["machine_specific_options"],
            config["system_keymap"],
        )
        # Do not reset if only the keymap changed.
        if (
            self._machine_params is None
            or self._machine_params.type != machine_params.type
            or self._machine_params.options != machine_params.options
        ):
            reset_machine = True
        if reset_machine:
            if self._machine is not None:
                self._machine.stop_capture()
                self._machine = None
            machine_class = registry.get_plugin("machine", machine_params.type).obj
            log.info("setting machine: %s", machine_params.type)
            self._machine = machine_class(machine_params.options)
            self._machine.set_suppression(self._is_running)
            self._machine.add_state_callback(self._machine_state_callback)
            self._machine.add_stroke_callback(self._machine_stroke_callback)
            self._machine_params = machine_params
            update_keymap = True
            start_machine = True
        elif self._machine is not None:
            update_keymap = "system_keymap" in config_update
        if update_keymap:
            machine_keymap = config["system_keymap"]
            if machine_keymap is not None:
                self._machine.set_keymap(machine_keymap)
        if start_machine:
            self._machine.start_capture()
        # Update running extensions.
        enabled_extensions = config["enabled_extensions"]
        running_extensions = set(self._running_extensions)
        self._stop_extensions(running_extensions - enabled_extensions)
        self._start_extensions(enabled_extensions - running_extensions)
        # Trigger `config_changed` hook.
        if config_update:
            self._trigger_hook("config_changed", config_update)
        # Update dictionaries.
        config_dictionaries = OrderedDict((d.path, d) for d in config["dictionaries"])
        copy_default_dictionaries(config_dictionaries.keys())
        # Start by unloading outdated dictionaries.
        self._dictionaries_manager.unload_outdated()
        self._set_dictionaries(
            [
                d
                for d in self._dictionaries.dicts
                if d.path in config_dictionaries
                and d.path in self._dictionaries_manager
            ]
        )
        # And then (re)load all dictionaries.
        dictionaries = []
        for d in self._dictionaries_manager.load(config_dictionaries.keys()):
            if isinstance(d, ErroredDictionary):
                # Only show an error if it's new.
                if d != self._dictionaries.get(d.path):
                    log.error(
                        "loading dictionary `%s` failed: %s",
                        shorten_path(d.path),
                        str(d.exception),
                    )
            d.enabled = config_dictionaries[d.path].enabled
            dictionaries.append(d)
        self._set_dictionaries(dictionaries)

    def _start_extensions(self, extension_list):
        for extension_name in extension_list:
            log.info("starting `%s` extension", extension_name)
            try:
                extension = registry.get_plugin("extension", extension_name).obj(self)
            except KeyError:
                # Plugin not installed, skip.
                continue
            try:
                extension.start()
            except Exception:
                log.error(
                    "initializing extension `%s` failed", extension_name, exc_info=True
                )
            else:
                self._running_extensions[extension_name] = extension

    def _stop_extensions(self, extension_list):
        for extension_name in list(extension_list):
            log.info("stopping `%s` extension", extension_name)
            extension = self._running_extensions.pop(extension_name)
            extension.stop()
            del extension

    def _quit(self, code):
        self._stop()
        self.code = code
        self._trigger_hook("quit")
        return True

    def _toggle_output(self):
        self._set_output(not self._is_running)

    def _set_output(self, enabled):
        if enabled == self._is_running:
            return
        self._is_running = enabled
        if enabled:
            self._translator.set_state(self._running_state)
        else:
            self._translator.clear_state()
        if self._machine is not None:
            self._machine.set_suppression(enabled)
        self._trigger_hook("output_changed", enabled)

    def _machine_state_callback(self, machine_state):
        self._same_thread_hook(self._on_machine_state_changed, machine_state)

    def _machine_stroke_callback(self, steno_keys):
        self._same_thread_hook(self._on_stroked, steno_keys)

    @with_lock
    def _on_machine_state_changed(self, machine_state):
        assert machine_state is not None
        self._machine_state = machine_state
        self._trigger_hook(
            "machine_state_changed", self._machine_params.type, machine_state
        )

    def _consume_engine_command(self, command, force=False):
        # The first commands can be used whether plover has output enabled or not.
        command_name, *command_args = command.split(":", 1)
        command_name = command_name.lower()
        if command_name == "resume":
            self._set_output(True)
            return True
        elif command_name == "toggle":
            self._toggle_output()
            return True
        elif command_name == "quit":
            self.quit()
            return True
        if not force and not self._is_running:
            return False
        # These commands can only be run when plover has output enabled.
        if command_name == "suspend":
            self._set_output(False)
        elif command_name == "configure":
            self._trigger_hook("configure")
        elif command_name == "focus":
            self._trigger_hook("focus")
        elif command_name == "add_translation":
            self._trigger_hook("add_translation")
        elif command_name == "lookup":
            self._trigger_hook("lookup")
        elif command_name == "suggestions":
            self._trigger_hook("suggestions")
        else:
            command_fn = registry.get_plugin("command", command_name).obj
            command_fn(self, command_args[0] if command_args else "")
        return False

    def _execute_engine_command(self, command, force=False):
        self._consume_engine_command(command, force=force)
        return False

    def _on_stroked(self, steno_keys):
        stroke = Stroke(steno_keys)
        log.stroke(stroke)
        self._translator.translate(stroke)
        self._trigger_hook("stroked", stroke)

    def _on_translated(self, old, new):
        if not self._is_running:
            return
        self._trigger_hook("translated", old, new)

    def _send_backspaces(self, b):
        if not self._is_running:
            return
        self._keyboard_emulation.send_backspaces(b)
        self._trigger_hook("send_backspaces", b)

    def _send_string(self, s):
        if not self._is_running:
            return
        self._keyboard_emulation.send_string(s)
        self._trigger_hook("send_string", s)

    def _send_key_combination(self, c):
        if not self._is_running:
            return
        self._keyboard_emulation.send_key_combination(c)
        self._trigger_hook("send_key_combination", c)

    def _send_engine_command(self, command):
        suppress = not self._is_running
        suppress &= self._consume_engine_command(command)
        if suppress:
            self._machine.suppress_last_stroke(self._keyboard_emulation.send_backspaces)

    def toggle_output(self):
        self._same_thread_hook(self._toggle_output)

    def set_output(self, enabled):
        self._same_thread_hook(self._set_output, enabled)

    @property
    @with_lock
    def machine_state(self):
        return self._machine_state

    @property
    @with_lock
    def output(self):
        return self._is_running

    @output.setter
    def output(self, enabled):
        self._same_thread_hook(self._set_output, enabled)

    @property
    @with_lock
    def config(self):
        return self._config.as_dict()

    @config.setter
    def config(self, update):
        self._same_thread_hook(self._update, config_update=update)

    @with_lock
    def __getitem__(self, setting):
        return self._config[setting]

    def __setitem__(self, setting, value):
        self.config = {setting: value}

    def reset_machine(self):
        self._same_thread_hook(self._update, reset_machine=True)

    def load_config(self):
        try:
            self._config.load()
        except Exception:
            log.error(
                "loading configuration failed, resetting to default", exc_info=True
            )
            self._config.clear()
            return False
        return True

    def start(self):
        self._same_thread_hook(self._start)

    def quit(self, code=0):
        # We need to go through the queue, even when already called
        # from the engine thread so _quit's return code does break
        # the thread out of its main loop.
        self._queue.put((self._quit, (code,), {}))

    def restart(self):
        self.quit(-1)

    def join(self):
        return self.code

    @with_lock
    def lookup(self, translation):
        return self._dictionaries.lookup(translation)

    @with_lock
    def raw_lookup(self, translation):
        return self._dictionaries.raw_lookup(translation)

    @with_lock
    def lookup_from_all(self, translation):
        return self._dictionaries.lookup_from_all(translation)

    @with_lock
    def raw_lookup_from_all(self, translation):
        return self._dictionaries.raw_lookup_from_all(translation)

    @with_lock
    def reverse_lookup(self, translation):
        matches = self._dictionaries.reverse_lookup(translation)
        return [] if matches is None else matches

    @with_lock
    def casereverse_lookup(self, translation):
        matches = self._dictionaries.casereverse_lookup(translation)
        return set() if matches is None else matches

    @with_lock
    def add_dictionary_filter(self, dictionary_filter):
        self._dictionaries.add_filter(dictionary_filter)

    @with_lock
    def remove_dictionary_filter(self, dictionary_filter):
        self._dictionaries.remove_filter(dictionary_filter)

    @with_lock
    def get_suggestions(self, translation):
        return Suggestions(self._dictionaries).find(translation)

    @property
    @with_lock
    def translator_state(self):
        return self._translator.get_state()

    @translator_state.setter
    @with_lock
    def translator_state(self, state):
        self._translator.set_state(state)

    @with_lock
    def clear_translator_state(self, undo=False):
        if undo:
            state = self._translator.get_state()
            if state.translations:
                self._formatter.format(state.translations, (), None)
        self._translator.clear_state()

    @property
    @with_lock
    def starting_stroke_state(self):
        return StartingStrokeState(
            self._formatter.start_attached,
            self._formatter.start_capitalized,
            self._formatter.space_char,
        )

    @starting_stroke_state.setter
    @with_lock
    def starting_stroke_state(self, state):
        self._formatter.start_attached = state.attach
        self._formatter.start_capitalized = state.capitalize
        self._formatter.space_char = state.space_char

    @with_lock
    def add_translation(self, strokes, translation, dictionary_path=None):
        if dictionary_path is None:
            dictionary_path = self._dictionaries.first_writable().path
        self._dictionaries.set(strokes, translation, path=dictionary_path)
        self._dictionaries.save(path_list=(dictionary_path,))

    @property
    @with_lock
    def dictionaries(self):
        return self._dictionaries

    # Hooks.

    def _trigger_hook(self, hook, *args, **kwargs):
        for callback in self._hooks[hook]:
            try:
                callback(*args, **kwargs)
            except Exception:
                log.error("hook %r callback %r failed", hook, callback, exc_info=True)

    @with_lock
    def hook_connect(self, hook, callback):
        self._hooks[hook].append(callback)

    @with_lock
    def hook_disconnect(self, hook, callback):
        self._hooks[hook].remove(callback)



from copy import copy
from pathlib import Path

from PySide6.QtCore import Qt, Signal, Slot
from PySide6.QtGui import (
    QIntValidator,
    QTextCharFormat,
    QTextFrameFormat,
    QTextListFormat,
    QTextCursor,
    QTextDocument,
)
from PySide6.QtWidgets import (
    QGroupBox,
    QStyle,
    QStyledItemDelegate,
    QToolTip,
)

from serial import Serial
from serial.tools.list_ports import comports

from plover import _
from plover.oslayer.serial import patch_ports_info

from plover.gui_qt.config_keyboard_widget_ui import Ui_KeyboardWidget
from plover.gui_qt.config_serial_widget_ui import Ui_SerialWidget
from plover.gui_qt.config_plover_hid_widget_ui import Ui_PloverHidWidget


def serial_port_details(port_info):
    parts = []
    global_ignore = {None, "n/a", Path(port_info.device).name}
    local_ignore = set(global_ignore)
    for attr, fmt in (
        ("product", _("product: {value}")),
        ("manufacturer", _("manufacturer: {value}")),
        ("serial_number", _("serial number: {value}")),
    ):
        value = getattr(port_info, attr)
        if value not in global_ignore:
            parts.append(fmt.format(value=value))
            local_ignore.add(value)
    description = getattr(port_info, "description")
    if description not in local_ignore:
        parts.insert(0, _("description: {value}").format(value=description))
    if not parts:
        return None
    return parts


class SerialOption(QGroupBox, Ui_SerialWidget):
    class PortDelegate(QStyledItemDelegate):
        def __init__(self):
            super().__init__()
            self._doc = QTextDocument()
            doc_margin = self._doc.documentMargin()
            self._doc.setIndentWidth(doc_margin * 3)
            background = QToolTip.palette().toolTipBase()
            foreground = QToolTip.palette().toolTipText()
            self._device_format = QTextCharFormat()
            self._details_char_format = QTextCharFormat()
            self._details_char_format.setFont(QToolTip.font())
            self._details_char_format.setBackground(background)
            self._details_char_format.setForeground(foreground)
            self._details_frame_format = QTextFrameFormat()
            self._details_frame_format.setBackground(background)
            self._details_frame_format.setForeground(foreground)
            self._details_frame_format.setTopMargin(doc_margin)
            self._details_frame_format.setBottomMargin(-3 * doc_margin)
            self._details_frame_format.setBorderStyle(
                QTextFrameFormat.BorderStyle.BorderStyle_Solid
            )
            self._details_frame_format.setBorder(doc_margin / 2)
            self._details_frame_format.setPadding(doc_margin)
            self._details_list_format = QTextListFormat()
            self._details_list_format.setStyle(QTextListFormat.Style.ListSquare)

        def _format_port(self, index):
            self._doc.clear()
            cursor = QTextCursor(self._doc)
            cursor.setCharFormat(self._device_format)
            port_info = index.data(Qt.ItemDataRole.UserRole)
            if port_info is None:
                cursor.insertText(index.data(Qt.ItemDataRole.DisplayRole))
                return
            cursor.insertText(port_info.device)
            details = serial_port_details(port_info)
            if not details:
                return
            cursor.insertFrame(self._details_frame_format)
            details_list = cursor.createList(self._details_list_format)
            for n, part in enumerate(details):
                if n:
                    cursor.insertBlock()
                cursor.insertText(part, self._details_char_format)

        def paint(self, painter, option, index):
            painter.save()
            if option.state & QStyle.StateFlag.State_Selected:
                painter.fillRect(option.rect, option.palette.highlight())
                text_color = option.palette.highlightedText()
            else:
                text_color = option.palette.text()
            self._device_format.setForeground(text_color)
            doc_margin = self._doc.documentMargin()
            self._details_frame_format.setWidth(option.rect.width() - doc_margin * 2)
            self._format_port(index)
            painter.translate(option.rect.topLeft())
            self._doc.drawContents(painter)
            painter.restore()

        def sizeHint(self, option, index):
            self._format_port(index)
            return self._doc.size().toSize()

    valueChanged = Signal(object)

    def __init__(self):
        super().__init__()
        self.setupUi(self)
        self.port.setItemDelegate(self.PortDelegate())
        self._value = {}

    def setValue(self, value):
        self._value = copy(value)
        self.scan()
        port = value["port"]
        if port is not None and port != "None":
            port_index = self.port.findText(port)
            if port_index != -1:
                self.port.setCurrentIndex(port_index)
            else:
                self.port.setCurrentText(port)
        self.baudrate.addItems(map(str, Serial.BAUDRATES))
        self.baudrate.setCurrentText(str(value["baudrate"]))
        self.bytesize.addItems(map(str, Serial.BYTESIZES))
        self.bytesize.setCurrentText(str(value["bytesize"]))
        self.parity.addItems(Serial.PARITIES)
        self.parity.setCurrentText(value["parity"])
        self.stopbits.addItems(map(str, Serial.STOPBITS))
        self.stopbits.setCurrentText(str(value["stopbits"]))
        timeout = value["timeout"]
        if timeout is None:
            self.use_timeout.setChecked(False)
            self.timeout.setValue(0.0)
            self.timeout.setEnabled(False)
        else:
            self.use_timeout.setChecked(True)
            self.timeout.setValue(timeout)
            self.timeout.setEnabled(True)
        for setting in ("xonxoff", "rtscts"):
            widget = getattr(self, setting)
            if setting in value:
                widget.setChecked(value[setting])
            else:
                widget.setEnabled(False)

    def _update(self, field, value):
        self._value[field] = value
        self.valueChanged.emit(self._value)

    def scan(self):
        self.port.clear()
        for port_info in sorted(patch_ports_info(comports())):
            self.port.addItem(port_info.device, port_info)

    @Slot(str)
    def update_port(self, value):
        self._update("port", value)

    @Slot(str)
    def update_baudrate(self, value):
        self._update("baudrate", int(value))

    @Slot(str)
    def update_bytesize(self, value):
        self._update("baudrate", int(value))

    @Slot(str)
    def update_parity(self, value):
        self._update("parity", value)

    def update_stopbits(self, value):
        self._update("stopbits", float(value))

    @Slot(float)
    def update_timeout(self, value):
        self._update("timeout", value)

    @Slot(bool)
    def update_use_timeout(self, value):
        if value:
            timeout = self.timeout.value()
        else:
            timeout = None
        self.timeout.setEnabled(value)
        self._update("timeout", timeout)

    @Slot(bool)
    def update_xonxoff(self, value):
        self._update("xonxoff", value)

    @Slot(bool)
    def update_rtscts(self, value):
        self._update("rtscts", value)


class KeyboardOption(QGroupBox, Ui_KeyboardWidget):
    valueChanged = Signal(object)

    def __init__(self):
        super().__init__()
        self.setupUi(self)
        self.arpeggiate.setToolTip(
            _(
                "Arpeggiate allows using non-NKRO keyboards.\n"
                "\n"
                "Each key can be pressed separately and the\n"
                "space bar is pressed to send the stroke."
            )
        )
        self._value = {}

    def setValue(self, value):
        self._value = copy(value)
        self.arpeggiate.setChecked(value["arpeggiate"])
        self.first_up_chord_send.setChecked(value["first_up_chord_send"])

    @Slot(bool)
    def update_arpeggiate(self, value):
        self._value["arpeggiate"] = value
        self.valueChanged.emit(self._value)

    @Slot(bool)
    def update_first_up_chord_send(self, value):
        self._value["first_up_chord_send"] = value
        self.valueChanged.emit(self._value)


class PloverHidOption(QGroupBox, Ui_PloverHidWidget):
    valueChanged = Signal(object)

    def __init__(self):
        super().__init__()
        self.setupUi(self)
        self._value = {}
        self.repeat_delay_ms.setValidator(QIntValidator(10, 10000, self))
        self.repeat_interval_ms.setValidator(QIntValidator(10, 10000, self))
        self.device_scan_interval_ms.setValidator(QIntValidator(250, 100000, self))

    def setValue(self, value):
        self._value = copy(value)
        self.first_up_chord_send.setChecked(value["first_up_chord_send"])
        self.double_tap_repeat.setChecked(value["double_tap_repeat"])
        self.repeat_delay_ms.setText(str(value["repeat_delay_ms"]))
        self.repeat_interval_ms.setText(str(value["repeat_interval_ms"]))
        self.device_scan_interval_ms.setText(
            str(value.get("device_scan_interval_ms", 1000))
        )

    @Slot(bool)
    def update_first_up_chord_send(self, value):
        self._value["first_up_chord_send"] = value
        self.valueChanged.emit(self._value)

    @Slot(bool)
    def update_double_tap_repeat(self, value):
        self._value["double_tap_repeat"] = value
        self.valueChanged.emit(self._value)

    @Slot(str)
    def update_repeat_delay_ms(self, text):
        if text.isdigit():
            self._value["repeat_delay_ms"] = int(text)
            self.valueChanged.emit(self._value)

    @Slot(str)
    def update_repeat_interval_ms(self, text):
        if text.isdigit():
            self._value["repeat_interval_ms"] = int(text)
            self.valueChanged.emit(self._value)

    @Slot(str)
    def update_device_scan_interval_ms(self, text):
        if text.isdigit():
            self._value["device_scan_interval_ms"] = int(text)
            self.valueChanged.emit(self._value)



import os
import gettext

from plover.oslayer.config import CONFIG_DIR, PLATFORM
from plover.oslayer.i18n import get_system_language
from plover.resource import ASSET_SCHEME, resource_filename


def get_language():
    # Give priority to LANGUAGE environment variable.
    lang = os.environ.get("LANGUAGE")
    if lang is not None:
        return lang
    # Try to get system language.
    lang = get_system_language()
    if lang is not None:
        return lang
    # Fallback to English.
    return "en"


def get_locale_dir(package, resource_dir):
    locale_dir = os.path.join(CONFIG_DIR, "messages")
    if gettext.find(package, locale_dir):
        return locale_dir
    return resource_filename(f"{ASSET_SCHEME}{package}:{resource_dir}")


class Translator:
    def __init__(self, package, resource_dir="messages", lang=None):
        self.package = package
        self.resource_dir = resource_dir
        if lang is None:
            lang = get_language()
        self.lang = lang

    @property
    def lang(self):
        return self._lang

    @lang.setter
    def lang(self, lang):
        self._lang = lang
        localedir = get_locale_dir(self.package, self.resource_dir)
        self._translation = gettext.translation(
            self.package, localedir=localedir, languages=[lang], fallback=True
        )
        self.gettext = self._translation.gettext
        self.ngettext = self._translation.ngettext

    def __call__(self, message):
        return self.gettext(message)

    def _(self, message):
        return message



from contextlib import contextmanager
import os

from PySide6.QtCore import (
    QAbstractListModel,
    QModelIndex,
    Qt,
    Signal,
    Slot,
)
from PySide6.QtGui import (
    QCursor,
    QIcon,
)
from PySide6.QtWidgets import (
    QFileDialog,
    QGroupBox,
    QMenu,
)

from plover import _
from plover.config import DictionaryConfig
from plover.dictionary.base import create_dictionary
from plover.engine import ErroredDictionary
from plover.misc import normalize_path
from plover.oslayer.config import CONFIG_DIR
from plover.registry import registry

from plover.gui_qt.dictionaries_widget_ui import Ui_DictionariesWidget
from plover.gui_qt.dictionary_editor import DictionaryEditor
from plover.gui_qt.utils import ToolBar


def _dictionary_formats(include_readonly=True):
    return {
        plugin.name
        for plugin in registry.list_plugins("dictionary")
        if include_readonly or not plugin.obj.readonly
    }


def _dictionary_filters(include_readonly=True):
    formats = sorted(_dictionary_formats(include_readonly=include_readonly))
    filters = ["*." + ext for ext in formats]
    # i18n: Widget: “DictionariesWidget”, file picker.
    filters = [_("Dictionaries ({extensions})").format(extensions=" ".join(filters))]
    filters.extend(
        # i18n: Widget: “DictionariesWidget”, file picker.
        _("{format} dictionaries ({extensions})").format(
            format=ext.strip(".").upper(),
            extensions="*." + ext,
        )
        for ext in formats
    )
    return ";; ".join(filters)


@contextmanager
def _new_dictionary(filename):
    try:
        d = create_dictionary(filename, threaded_save=False)
        yield d
        d.save()
    except Exception as e:
        raise Exception("creating dictionary %s failed. %s" % (filename, e)) from e


class DictionariesModel(QAbstractListModel):
    class DictionaryItem:
        __slots__ = "row path enabled short_path _loaded state".split()

        def __init__(self, row, config, loaded=None):
            self.row = row
            self.path = config.path
            self.enabled = config.enabled
            self.short_path = config.short_path
            self.loaded = loaded

        @property
        def loaded(self):
            return self._loaded

        @loaded.setter
        def loaded(self, loaded):
            if loaded is None:
                state = "loading"
            elif isinstance(loaded, ErroredDictionary):
                state = "error"
            elif loaded.readonly:
                state = "readonly"
            else:
                state = "normal"
            self.state = state
            self._loaded = loaded

        @property
        def config(self):
            return DictionaryConfig(self.path, self.enabled)

        @property
        def is_loaded(self):
            return self.state not in {"loading", "error"}

    SUPPORTED_ROLES = {
        Qt.ItemDataRole.AccessibleTextRole,
        Qt.ItemDataRole.CheckStateRole,
        Qt.ItemDataRole.DecorationRole,
        Qt.ItemDataRole.DisplayRole,
        Qt.ItemDataRole.ToolTipRole,
    }

    FLAGS = (
        Qt.ItemFlag.ItemIsEnabled
        | Qt.ItemFlag.ItemIsSelectable
        | Qt.ItemFlag.ItemIsUserCheckable
        | Qt.ItemFlag.ItemIsDragEnabled
    )

    has_undo_changed = Signal(bool)

    def __init__(self, engine, icons, max_undo=20):
        super().__init__()
        self._engine = engine
        self._favorite = None
        self._from_path = {}
        self._from_row = []
        self._reverse_order = False
        self._undo_stack = []
        self._max_undo = max_undo
        self._icons = icons
        with engine:
            config = engine.config
            engine.signal_connect("config_changed", self._on_config_changed)
            engine.signal_connect("dictionaries_loaded", self._on_dictionaries_loaded)
            engine.signal_connect(
                "dictionary_state_changed", self._on_dictionary_state_changed
            )
            self._reset_items(
                config["dictionaries"],
                config["classic_dictionaries_display_order"],
                backup=False,
                publish=False,
            )

    @property
    def _config(self):
        item_list = self._from_row
        if self._reverse_order:
            item_list = reversed(item_list)
        return [item.config for item in item_list]

    def _updated_rows(self, row_list):
        for row in row_list:
            index = self.index(row)
            self.dataChanged.emit(index, index, self.SUPPORTED_ROLES)

    def _backup_config(self):
        config = self._config
        assert not self._undo_stack or config != self._undo_stack[-1]
        self._undo_stack.append(config)
        if len(self._undo_stack) == 1:
            self.has_undo_changed.emit(True)
        elif len(self._undo_stack) > self._max_undo:
            self._undo_stack.pop(0)

    def _publish_config(self, config):
        self._engine.config = {"dictionaries": config}

    def _update_favorite(self):
        item_list = self._from_row
        if self._reverse_order:
            item_list = reversed(item_list)
        old, new = self._favorite, None
        for item in item_list:
            if item.enabled and item.state == "normal":
                new = item
                break
        if new is old:
            return set()
        self._favorite = new
        return {favorite.row for favorite in filter(None, (old, new))}

    def _reset_items(self, config, reverse_order=None, backup=True, publish=True):
        if backup:
            self._backup_config()
        if reverse_order is None:
            reverse_order = self._reverse_order
        self.layoutAboutToBeChanged.emit()
        old_persistent_indexes = self.persistentIndexList()
        assert all(index.isValid() for index in old_persistent_indexes)
        old_persistent_items = [
            self._from_row[index.row()] for index in old_persistent_indexes
        ]
        from_row = []
        from_path = {}
        for row, d in enumerate(reversed(config) if reverse_order else config):
            item = self._from_path.get(d.path)
            if item is None:
                item = self.DictionaryItem(row, d)
            else:
                item.row = row
                item.enabled = d.enabled
            assert d.path not in from_path
            from_path[d.path] = item
            from_row.append(item)
        self._reverse_order = reverse_order
        self._from_path = from_path
        self._from_row = from_row
        self._update_favorite()
        new_persistent_indexes = []
        for old_item in old_persistent_items:
            new_item = self._from_path.get(old_item.path)
            if new_item is None:
                new_index = QModelIndex()
            else:
                new_index = self.index(new_item.row)
            new_persistent_indexes.append(new_index)
        self.changePersistentIndexList(old_persistent_indexes, new_persistent_indexes)
        self.layoutChanged.emit()
        if publish:
            self._publish_config(config)

    def _on_config_changed(self, config_update):
        config = config_update.get("dictionaries")
        reverse_order = config_update.get("classic_dictionaries_display_order")
        noop = True
        if reverse_order is not None:
            noop = reverse_order == self._reverse_order
        if config is not None:
            noop = noop and config == self._config
        if noop:
            return
        if config is None:
            config = self._config
        else:
            if self._undo_stack:
                self.has_undo_changed.emit(False)
            self._undo_stack.clear()
        self._reset_items(config, reverse_order, backup=False, publish=False)

    def _on_dictionaries_loaded(self, loaded_dictionaries):
        updated_rows = set()
        for item in self._from_row:
            loaded = loaded_dictionaries.get(item.path)
            if loaded == item.loaded:
                continue
            item.loaded = loaded
            updated_rows.add(item.row)
        if not updated_rows:
            return
        updated_rows.update(self._update_favorite())
        self._updated_rows(updated_rows)

    def _on_dictionary_state_changed(self, filename, d):
        [item] = [item for item in self._from_row if item.path == filename]
        if item.loaded != d:
            item.loaded = d
            self._updated_rows([item.row])

    def _move(self, index_list, step):
        row_list = sorted(self._normalized_row_list(index_list))
        if not row_list:
            return
        old_path_list = [item.path for item in self._from_row]
        new_path_list = old_path_list[:]
        if step > 0:
            row_list = reversed(row_list)
            row_limit = len(self._from_row) - 1
            update_row = min
        else:
            row_limit = 0
            update_row = max
        for old_row in row_list:
            new_row = update_row(old_row + step, row_limit)
            new_path_list.insert(new_row, new_path_list.pop(old_row))
            row_limit = new_row - step
        if old_path_list == new_path_list:
            return
        if self._reverse_order:
            new_path_list = reversed(new_path_list)
        config = [self._from_path[path].config for path in new_path_list]
        self._reset_items(config)

    @staticmethod
    def _normalized_path_list(path_list):
        return list(dict.fromkeys(map(normalize_path, path_list)))

    @staticmethod
    def _normalized_row_list(index_list):
        return list(
            dict.fromkeys(index.row() for index in index_list if index.isValid())
        )

    def _insert(self, dest_row, path_list):
        old_path_list = [item.path for item in self._from_row]
        new_path_list = (
            [p for p in old_path_list[:dest_row] if p not in path_list]
            + path_list
            + [p for p in old_path_list[dest_row:] if p not in path_list]
        )
        if new_path_list == old_path_list:
            return
        if self._reverse_order:
            new_path_list = reversed(new_path_list)
        config = [
            self._from_path[path].config
            if path in self._from_path
            else DictionaryConfig(path)
            for path in new_path_list
        ]
        self._reset_items(config)

    def add(self, path_list):
        new_path_list = self._normalized_path_list(
            path for path in path_list if path not in self._from_path
        )
        if new_path_list:
            # Add with highest priority.
            if self._reverse_order:
                dest_row = len(self._from_path)
                new_path_list = reversed(new_path_list)
            else:
                dest_row = 0
            self._insert(dest_row, new_path_list)
        elif path_list:
            # Trigger a reload, just in case.
            self._engine.config = {}

    def iter_loaded(self, index_list):
        row_list = sorted(self._normalized_row_list(index_list))
        if self._reverse_order:
            row_list = reversed(row_list)
        for row in row_list:
            item = self._from_row[row]
            if item.is_loaded:
                yield item.loaded

    def insert(self, index, path_list):
        # Insert at the end if `index` is not valid.
        row = index.row() if index.isValid() else len(self._from_row)
        self._insert(row, self._normalized_path_list(path_list))

    def move(self, dst_index, src_index_list):
        self.insert(
            dst_index,
            [
                self._from_row[row].path
                for row in self._normalized_row_list(src_index_list)
            ],
        )

    def move_down(self, index_list):
        self._move(index_list, +1)

    def move_up(self, index_list):
        self._move(index_list, -1)

    def remove(self, index_list):
        row_set = self._normalized_row_list(index_list)
        if not row_set:
            return
        config = [item.config for item in self._from_row if item.row not in row_set]
        self._reset_items(config)

    def undo(self):
        config = self._undo_stack.pop()
        if not self._undo_stack:
            self.has_undo_changed.emit(False)
        self._reset_items(config, backup=False)

    # Model API.

    def rowCount(self, parent=QModelIndex()):
        return 0 if parent.isValid() else len(self._from_row)

    def flags(self, index):
        return self.FLAGS if index.isValid() else Qt.ItemFlag.NoItemFlags

    def data(self, index, role):
        if not index.isValid() or role not in self.SUPPORTED_ROLES:
            return None
        d = self._from_row[index.row()]
        if role == Qt.ItemDataRole.DisplayRole:
            return d.short_path
        if role == Qt.ItemDataRole.CheckStateRole:
            return Qt.CheckState.Checked if d.enabled else Qt.CheckState.Unchecked
        if role == Qt.ItemDataRole.AccessibleTextRole:
            accessible_text = [d.short_path]
            if not d.enabled:
                # i18n: Widget: “DictionariesWidget”, accessible text.
                accessible_text.append(_("disabled"))
            if d is self._favorite:
                # i18n: Widget: “DictionariesWidget”, accessible text.
                accessible_text.append(_("favorite"))
            elif d.state == "error":
                # i18n: Widget: “DictionariesWidget”, accessible text.
                accessible_text.append(
                    _("errored: {exception}.").format(exception=str(d.loaded.exception))
                )
            elif d.state == "loading":
                # i18n: Widget: “DictionariesWidget”, accessible text.
                accessible_text.append(_("loading"))
            elif d.state == "readonly":
                # i18n: Widget: “DictionariesWidget”, accessible text.
                accessible_text.append(_("read-only"))
            return ", ".join(accessible_text)
        if role == Qt.ItemDataRole.DecorationRole:
            return self._icons.get("favorite" if d is self._favorite else d.state)
        if role == Qt.ItemDataRole.ToolTipRole:
            # i18n: Widget: “DictionariesWidget”, tooltip.
            tooltip = [_("Full path: {path}.").format(path=d.config.path)]
            if d is self._favorite:
                # i18n: Widget: “DictionariesWidget”, tool tip.
                tooltip.append(_("This dictionary is marked as the favorite."))
            elif d.state == "loading":
                # i18n: Widget: “DictionariesWidget”, tool tip.
                tooltip.append(_("This dictionary is being loaded."))
            elif d.state == "error":
                # i18n: Widget: “DictionariesWidget”, tool tip.
                tooltip.append(
                    _("Loading this dictionary failed: {exception}.").format(
                        exception=str(d.loaded.exception)
                    )
                )
            elif d.state == "readonly":
                # i18n: Widget: “DictionariesWidget”, tool tip.
                tooltip.append(_("This dictionary is read-only."))
            return "\n\n".join(tooltip)
        return None

    def setData(self, index, value, role):
        if not index.isValid() or role != Qt.ItemDataRole.CheckStateRole.value:
            return False
        if value == Qt.CheckState.Checked.value:
            enabled = True
        elif value == Qt.CheckState.Unchecked.value:
            enabled = False
        else:
            return False
        d = self._from_row[index.row()]
        if d.enabled == enabled:
            return False
        self._backup_config()
        d.enabled = enabled
        self._updated_rows({d.row} | self._update_favorite())
        self._publish_config(self._config)
        return True


class DictionariesWidget(QGroupBox, Ui_DictionariesWidget):
    signal_add_translation = Signal(str)

    # TODO what to do with parent?
    def __init__(self, parent=None):
        super().__init__()
        self.setupUi(self)
        self._setup = False
        self._engine = None
        self._model = None
        # The save/open/new dialogs will open on that directory.
        self._file_dialogs_directory = CONFIG_DIR
        # Start with all actions disabled (until `setup` is called).
        for action in (
            self.action_AddDictionaries,
            self.action_AddTranslation,
            self.action_EditDictionaries,
            self.action_MoveDictionariesDown,
            self.action_MoveDictionariesUp,
            self.action_RemoveDictionaries,
            self.action_SaveDictionaries,
            self.action_Undo,
        ):
            action.setEnabled(False)
        # Toolbar.
        self.layout().addWidget(
            ToolBar(
                self.action_Undo,
                None,
                self.action_EditDictionaries,
                self.action_RemoveDictionaries,
                self.action_AddDictionaries,
                self.action_AddTranslation,
                None,
                self.action_MoveDictionariesUp,
                self.action_MoveDictionariesDown,
            )
        )
        # Add menu.
        self.menu_AddDictionaries = QMenu(self.action_AddDictionaries.text())
        self.menu_AddDictionaries.setIcon(self.action_AddDictionaries.icon())
        self.menu_AddDictionaries.addAction(self.action_OpenDictionaries)
        self.menu_AddDictionaries.addAction(self.action_CreateDictionary)
        self.action_AddDictionaries.setMenu(self.menu_AddDictionaries)
        # Save menu.
        self.menu_SaveDictionaries = QMenu(self.action_SaveDictionaries.text())
        self.menu_SaveDictionaries.setIcon(self.action_SaveDictionaries.icon())
        self.menu_SaveDictionaries.addAction(self.action_CopyDictionaries)
        self.menu_SaveDictionaries.addAction(self.action_MergeDictionaries)
        self.view.dragEnterEvent = self._drag_enter_event
        self.view.dragMoveEvent = self._drag_move_event
        self.view.dropEvent = self._drag_drop_event
        self.setFocusProxy(self.view)
        # Edit context menu.
        edit_menu = QMenu()
        edit_menu.addAction(self.action_Undo)
        edit_menu.addSeparator()
        edit_menu.addMenu(self.menu_AddDictionaries)
        edit_menu.addAction(self.action_EditDictionaries)
        edit_menu.addMenu(self.menu_SaveDictionaries)
        edit_menu.addAction(self.action_RemoveDictionaries)
        edit_menu.addSeparator()
        edit_menu.addAction(self.action_MoveDictionariesUp)
        edit_menu.addAction(self.action_MoveDictionariesDown)
        self.view.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.view.customContextMenuRequested.connect(
            lambda p: edit_menu.exec(self.view.mapToGlobal(p))
        )
        self.edit_menu = edit_menu

    def setup(self, engine):
        assert not self._setup
        self._engine = engine
        self._model = DictionariesModel(
            engine,
            {
                name: QIcon(":resources/dictionary_%s.svg" % name)
                for name in "favorite loading error readonly normal".split()
            },
        )
        self._model.has_undo_changed.connect(self.toggle_undo_action)
        self.view.setModel(self._model)
        self.view.selectionModel().selectionChanged.connect(
            self.handle_selection_change
        )
        for action in (
            self.action_AddDictionaries,
            self.action_AddTranslation,
        ):
            action.setEnabled(True)
        self._setup = True

    @property
    def _selected(self):
        return sorted(self.view.selectedIndexes(), key=lambda index: index.row())

    def _drag_accept(self, event):
        accepted = False
        if event.source() is self.view:
            accepted = True
        elif event.mimeData().hasUrls():
            # Only allow a list of supported local files.
            for url in event.mimeData().urls():
                if not url.isLocalFile():
                    break
                filename = url.toLocalFile()
                extension = os.path.splitext(filename)[1].lower()[1:]
                if extension not in _dictionary_formats():
                    break
            else:
                accepted = True
        if accepted:
            event.accept()
        return accepted

    def _drag_enter_event(self, event):
        self._drag_accept(event)

    def _drag_move_event(self, event):
        self._drag_accept(event)

    def _drag_drop_event(self, event):
        if not self._drag_accept(event):
            return
        index = self.view.indexAt(event.pos())
        if event.source() == self.view:
            self._model.move(index, self._selected)
        else:
            path_list = [url.toLocalFile() for url in event.mimeData().urls()]
            self._model.insert(index, path_list)

    def _get_dictionary_save_name(
        self, title, default_name=None, default_extensions=(), initial_filename=None
    ):
        if default_name is not None:
            # Default to a writable dictionary format.
            writable_extensions = set(_dictionary_formats(include_readonly=False))
            default_name += "." + next(
                (e for e in default_extensions if e in writable_extensions), "json"
            )
            default_name = os.path.join(self._file_dialogs_directory, default_name)
        else:
            default_name = self._file_dialogs_directory
        default_name = normalize_path(default_name)
        new_filename = QFileDialog.getSaveFileName(
            parent=self,
            caption=title,
            dir=default_name,
            filter=_dictionary_filters(include_readonly=False),
        )[0]
        if not new_filename:
            return None
        new_filename = normalize_path(new_filename)
        self._file_dialogs_directory = os.path.dirname(new_filename)
        if new_filename == initial_filename:
            return None
        return new_filename

    def _edit_dictionaries(self, index_list):
        path_list = [d.path for d in self._model.iter_loaded(index_list)]
        if not path_list:
            return
        editor = DictionaryEditor(self._engine, path_list)
        editor.exec()

    def _copy_dictionaries(self, dictionaries):
        need_reload = False
        # i18n: Widget: “DictionariesWidget”, “save as copy” file picker.
        title_template = _("Save a copy of {name} as...")
        # i18n: Widget: “DictionariesWidget”, “save as copy” file picker.
        default_name_template = _("{name} - Copy")
        for original in dictionaries:
            title = title_template.format(name=os.path.basename(original.path))
            name, ext = os.path.splitext(os.path.basename(original.path))
            default_name = default_name_template.format(name=name)
            new_filename = self._get_dictionary_save_name(
                title, default_name, [ext[1:]], initial_filename=original.path
            )
            if new_filename is None:
                continue
            with _new_dictionary(new_filename) as copy:
                copy.update(original)
            need_reload = True
        return need_reload

    def _merge_dictionaries(self, dictionaries):
        names, exts = zip(
            *(os.path.splitext(os.path.basename(d.path)) for d in dictionaries)
        )
        default_name = " + ".join(names)
        default_exts = list(dict.fromkeys(e[1:] for e in exts))
        # i18n: Widget: “DictionariesWidget”, “save as merge” file picker.
        title = _("Merge {names} as...").format(names=default_name)
        new_filename = self._get_dictionary_save_name(title, default_name, default_exts)
        if new_filename is None:
            return False
        with _new_dictionary(new_filename) as merge:
            # Merge in reverse priority order, so higher
            # priority entries overwrite lower ones.
            for source in reversed(dictionaries):
                merge.update(source)
        return True

    def _save_dictionaries(self, merge=False):
        # Ignore dictionaries that are not loaded.
        dictionaries = list(self._model.iter_loaded(self._selected))
        if not dictionaries:
            return
        if merge:
            save_fn = self._merge_dictionaries
        else:
            save_fn = self._copy_dictionaries
        if save_fn(dictionaries):
            # This will trigger a reload of any modified dictionary.
            self._engine.config = {}

    @Slot()
    def open_open_dictionaries_dialog(self):
        new_filenames = QFileDialog.getOpenFileNames(
            # i18n: Widget: “DictionariesWidget”, “add” file picker.
            parent=self,
            caption=_("Load dictionaries"),
            dir=self._file_dialogs_directory,
            filter=_dictionary_filters(),
        )[0]
        if not new_filenames:
            return
        self._file_dialogs_directory = os.path.dirname(new_filenames[-1])
        self._model.add(new_filenames)

    @Slot()
    def open_create_new_dictionary_dialog(self):
        # i18n: Widget: “DictionariesWidget”, “new” file picker.
        new_filename = self._get_dictionary_save_name(_("Create dictionary"))
        if new_filename is None:
            return
        with _new_dictionary(new_filename):
            pass
        self._model.add([new_filename])

    @Slot()
    def copy_selected_dictionaries(self):
        self._save_dictionaries()

    @Slot()
    def merge_selected_dictionaries(self):
        self._save_dictionaries(merge=True)

    @Slot(QModelIndex)
    def activate_dictionary(self, index):
        self._edit_dictionaries([index])

    @Slot()
    def open_add_dictionaries_menu(self):
        self.menu_AddDictionaries.exec(QCursor.pos())

    @Slot()
    def add_translation(self):
        dictionary = next(self._model.iter_loaded([self.view.currentIndex()]), None)
        self.signal_add_translation.emit(
            None if dictionary is None else dictionary.path
        )

    @Slot()
    def edit_selected_dictionaries(self):
        self._edit_dictionaries(self._selected)

    @Slot(bool)
    def toggle_undo_action(self, available):
        self.action_Undo.setEnabled(available)

    @Slot()
    def move_selected_dictionaries_down(self):
        self._model.move_down(self._selected)

    @Slot()
    def move_selected_dictionaries_up(self):
        self._model.move_up(self._selected)

    @Slot()
    def remove_selected_dictionaries(self):
        self._model.remove(self._selected)

    @Slot()
    def handle_selection_change(self):
        selection = self._selected
        has_selection = bool(selection)
        for widget in (
            self.action_RemoveDictionaries,
            self.action_MoveDictionariesUp,
            self.action_MoveDictionariesDown,
        ):
            widget.setEnabled(has_selection)
        has_live_selection = next(self._model.iter_loaded(selection), None) is not None
        for widget in (
            self.action_EditDictionaries,
            self.action_SaveDictionaries,
            self.menu_SaveDictionaries,
        ):
            widget.setEnabled(has_live_selection)

    @Slot()
    def undo(self):
        self._model.undo()



#!/usr/bin/env python3

from collections import OrderedDict
from importlib.metadata import distributions, requires, PackageNotFoundError
from packaging.requirements import Requirement

from plover.registry import Registry


def sorted_requirements(requirements):
    return sorted(requirements, key=lambda r: str(r).lower())


# Find all available distributions.
all_requirements = [
    Requirement(f"{dist.metadata['Name']}=={dist.version}") for dist in distributions()
]

# Find Plover requirements.
plover_deps = set()
try:
    plover_requires = requires("plover") or []
    for req in plover_requires:
        plover_deps.add(Requirement(req))
except PackageNotFoundError:
    print("Plover is not installed.")
    plover_requires = []

# List requirements.
print("# plover")
for requirement in sorted_requirements(plover_deps):
    print(requirement)
print("# other")
for requirement in sorted_requirements(all_requirements):
    if requirement not in plover_deps:
        print(requirement)
print("# vim: ft=cfg commentstring=#\\ %s list")



from plover.steno import Stroke


def last_stroke(translator, stroke, cmdline):
    assert not cmdline
    # Repeat last stroke
    translations = translator.get_state().translations
    if not translations:
        return
    stroke = Stroke(translations[-1].strokes[-1].steno_keys)
    translator.translate_stroke(stroke)



from pathlib import Path
import inspect

import pytest

from plover.misc import normalize_path
from plover.resource import (
    resource_exists,
    resource_filename,
    resource_timestamp,
    resource_update,
)


@pytest.mark.parametrize(
    "resource, exists, filename",
    (
        # Relative filename.
        (Path(__file__).relative_to(Path.cwd()), True, None),
        # Relative directory.
        (Path(__file__).parent.relative_to(Path.cwd()), True, None),
        # Absolute filename.
        (__file__, True, None),
        # Absolute directory.
        (Path(__file__).parent, True, None),
        # Missing relative path.
        ("test/pouet", False, None),
        # Missing absolute path.
        (Path.cwd() / "test" / "pouet", False, None),
        # Asset filename.
        ("asset:plover:assets/user.json", True, "plover/assets/user.json"),
        # Asset directory.
        ("asset:plover:", True, "plover"),
        ("asset:plover:assets", True, "plover/assets"),
        # Missing asset.
        ("asset:plover:assets/pouet.json", False, "plover/assets/pouet.json"),
        # Invalid asset: missing package and path.
        ("asset:", ValueError, ValueError),
        # Invalid asset: missing path.
        ("asset:package", ValueError, ValueError),
        # Invalid asset: absolute resource path.
        ("asset:plover:/assets/user.json", ValueError, ValueError),
    ),
)
def test_resource(resource, exists, filename):
    resource = str(resource)
    if inspect.isclass(exists):
        exception = exists
        with pytest.raises(exception):
            resource_exists(resource)
        with pytest.raises(exception):
            resource_filename(resource)
        with pytest.raises(exception):
            resource_timestamp(resource)
        return
    assert resource_exists(resource) == exists
    if filename is None:
        filename = resource
    assert normalize_path(resource_filename(resource)) == normalize_path(filename)
    if exists:
        timestamp = Path(filename).stat().st_mtime
        assert resource_timestamp(resource) == timestamp
    else:
        with pytest.raises(FileNotFoundError):
            resource_timestamp(resource)


def test_resource_update(tmp_path):
    # Can't update assets.
    resource = "asset:plover:assets/pouet.json"
    resource_path = Path(resource_filename(resource))
    with pytest.raises(ValueError):
        with resource_update(resource):
            resource_path.write_bytes(b"contents")
    assert not resource_path.exists()
    # Don't update resource on exception (but still cleanup).
    resource = (tmp_path / "resource").resolve()
    exception_str = "Houston, we have a problem"
    with pytest.raises(Exception, match=exception_str):
        with resource_update(str(resource)) as tmpf:
            tmpf = Path(tmpf)
            tmpf.write_bytes(b"contents")
            raise Exception(exception_str)
    assert not resource.exists()
    assert not tmpf.exists()
    # Normal use.
    with resource_update(str(resource)) as tmpf:
        tmpf = Path(tmpf).resolve()
        # Temporary filename must be different.
        assert tmpf != resource
        # And must be empty.
        assert not tmpf.stat().st_size
        # Temporary file must be created in the same
        # directory as the target resource (so an
        # atomic rename can be used).
        assert tmpf.parent == resource.parent
        # Save something.
        tmpf.write_bytes(b"contents")
        st = tmpf.stat()
    assert resource.stat() == st
    assert resource.read_bytes() == b"contents"



import logging

from PySide6.QtCore import QObject, Signal

from plover import log


class NotificationHandler(QObject, logging.Handler):
    emitSignal = Signal(int, str)

    def __init__(self):
        super().__init__()
        self.setLevel(log.WARNING)
        self.setFormatter(
            log.NoExceptionTracebackFormatter("%(levelname)s: %(message)s")
        )

    def handle(self, record):
        level = record.levelno
        message = self.format(record)
        self.emitSignal.emit(level, message)



class CaptureOutput:
    def __init__(self):
        self.instructions = []
        self.text = ""

    def send_backspaces(self, n):
        assert n <= len(self.text)
        self.text = self.text[:-n]
        self.instructions.append(("b", n))

    def send_string(self, s):
        self.text += s
        self.instructions.append(("s", s))

    def send_key_combination(self, c):
        self.instructions.append(("c", c))

    def send_engine_command(self, c):
        self.instructions.append(("e", c))



from requests_futures import sessions
import requests_cache


class CachedSession(requests_cache.CachedSession):
    def __init__(self):
        super().__init__(backend="memory", expire_after=600)
        self.cache.delete(expired=True)


class CachedFuturesSession(sessions.FuturesSession):
    def __init__(self, session=None):
        if session is None:
            session = CachedSession()
        super().__init__(session=session, max_workers=4)



from plover.formatting import Case, apply_case


def meta_case(ctx, case):
    case = Case(case.lower())
    action = ctx.copy_last_action()
    action.next_case = case
    return action


def meta_retro_case(ctx, case):
    case = Case(case.lower())
    action = ctx.copy_last_action()
    action.prev_attach = True
    last_words = ctx.last_words(count=1)
    if last_words:
        action.prev_replace = last_words[0]
        action.text = apply_case(last_words[0], case)
    else:
        action.text = ""
    return action



# Copyright (c) 2011 Joshua Harlan Lifton.
# See LICENSE.txt for details.

"""Custom exceptions used by Plover.

The exceptions in this module are typically caught in the main GUI
loop and displayed to the user as an alert dialog.

"""


class InvalidConfigurationError(Exception):
    "Raised when there is something wrong in the configuration."

    pass



import pytest

from plover import system
from plover.machine.keyboard import Keyboard
from plover.machine.keymap import Keymap
from plover.oslayer.keyboardcontrol import KeyboardCapture

from .py37compat import mock


def send_input(capture, key_events):
    for evt in key_events.strip().split():
        if evt.startswith("+"):
            capture.key_down(evt[1:])
        elif evt.startswith("-"):
            capture.key_up(evt[1:])
        else:
            capture.key_down(evt)
            capture.key_up(evt)


@pytest.fixture
def capture():
    capture = mock.MagicMock(spec=KeyboardCapture)
    with mock.patch("plover.machine.keyboard.KeyboardCapture", new=lambda: capture):
        yield capture


@pytest.fixture(params=[{"arpeggiate": False, "first_up_chord_send": False}])
def machine(request, capture):
    machine = Keyboard(request.param)
    keymap = Keymap(Keyboard.KEYS_LAYOUT.split(), system.KEYS + Keyboard.ACTIONS)
    keymap.set_mappings(system.KEYMAPS["Keyboard"])
    machine.set_keymap(keymap)
    return machine


arpeggiate = pytest.mark.parametrize(
    "machine", [{"arpeggiate": True, "first_up_chord_send": False}], indirect=True
)
first_up_chord_send = pytest.mark.parametrize(
    "machine", [{"arpeggiate": False, "first_up_chord_send": True}], indirect=True
)
"""
These are decorators to be applied on test functions to modify the machine configuration.
Note that at the moment it's not possible to apply both at the same time.
"""


@pytest.fixture
def strokes(machine):
    strokes = []
    machine.add_stroke_callback(strokes.append)
    return strokes


def test_lifecycle(capture, machine, strokes):
    # Start machine.
    machine.start_capture()
    assert capture.mock_calls == [
        mock.call.start(),
        mock.call.suppress(()),
    ]
    capture.reset_mock()
    machine.set_suppression(True)
    suppressed_keys = dict(machine.keymap.get_bindings())
    del suppressed_keys["space"]
    assert strokes == []
    assert capture.mock_calls == [
        mock.call.suppress(suppressed_keys.keys()),
    ]
    # Trigger some strokes.
    capture.reset_mock()
    send_input(capture, "+a +h -a -h space w")
    assert strokes == [
        {"S-", "*"},
        {"T-"},
    ]
    assert capture.mock_calls == []
    # Stop machine.
    del strokes[:]
    machine.stop_capture()
    assert strokes == []
    assert capture.mock_calls == [
        mock.call.suppress(()),
        mock.call.cancel(),
    ]


def test_unfinished_stroke_1(capture, machine, strokes):
    machine.start_capture()
    send_input(capture, "+a +q -a")
    assert strokes == []


def test_unfinished_stroke_2(capture, machine, strokes):
    machine.start_capture()
    send_input(capture, "+a +r -a +a -r")
    assert strokes == []


@arpeggiate
def test_arpeggiate_1(capture, machine, strokes):
    machine.start_capture()
    send_input(capture, "a h space w")
    assert strokes == [{"S-", "*"}]


@arpeggiate
def test_arpeggiate_2(capture, machine, strokes):
    machine.start_capture()
    send_input(capture, "a +h +space -space -h w")
    assert strokes == [{"S-", "*"}]


@first_up_chord_send
def test_first_up_chord_send(capture, machine, strokes):
    machine.start_capture()
    send_input(capture, "+a +w +l -l +l")
    assert strokes == [{"S-", "T-", "-G"}]
    send_input(capture, "-l")
    assert strokes == [{"S-", "T-", "-G"}, {"S-", "T-", "-G"}]
    send_input(capture, "-a -w")
    assert strokes == [{"S-", "T-", "-G"}, {"S-", "T-", "-G"}]



# Copyright (c) 2011 Hesky Fisher
# See LICENSE.txt for details.

"Thread-based monitoring of a stenotype machine using the TX Bolt protocol."

import plover.machine.base

# In the TX Bolt protocol, there are four sets of keys grouped in
# order from left to right. Each byte represents all the keys that
# were pressed in that set. The first two bits indicate which set this
# byte represents. The next bits are set if the corresponding key was
# pressed for the stroke.

# 00XXXXXX 01XXXXXX 10XXXXXX 110XXXXX
#   HWPKTS   UE*OAR   GLBPRF    #ZDST

# The protocol uses variable length packets of one, two, three or four
# bytes. Only those bytes for which keys were pressed will be
# transmitted. The bytes arrive in order of the sets so it is clear
# when a new stroke starts. Also, if a key is pressed in an earlier
# set in one stroke and then a key is pressed only in a later set then
# there will be a zero byte to indicate that this is a new stroke. So,
# it is reliable to assume that a stroke ended when a lower set is
# seen. Additionally, if there is no activity then the machine will
# send a zero byte every few seconds.

# fmt: off
STENO_KEY_CHART = ("S-", "T-", "K-", "P-", "W-", "H-",  # 00
                   "R-", "A-", "O-", "*", "-E", "-U",   # 01
                   "-F", "-R", "-P", "-B", "-L", "-G",  # 10
                   "-T", "-S", "-D", "-Z", "#")         # 11
# fmt: on


class TxBolt(plover.machine.base.SerialStenotypeBase):
    """TX Bolt interface.

    This class implements the three methods necessary for a standard
    stenotype interface: start_capture, stop_capture, and
    add_callback.

    """

    KEYS_LAYOUT = """
        #  #  #  #  #  #  #  #  #  #
        S- T- P- H- * -F -P -L -T -D
        S- K- W- R- * -R -B -G -S -Z
              A- O-   -E -U
    """

    def __init__(self, params):
        super().__init__(params)
        self._reset_stroke_state()

    def _reset_stroke_state(self):
        self._pressed_keys = []
        self._last_key_set = 0

    def _finish_stroke(self):
        steno_keys = self.keymap.keys_to_actions(self._pressed_keys)
        if steno_keys:
            self._notify(steno_keys)
        self._reset_stroke_state()

    def run(self):
        """Overrides base class run method. Do not call directly."""
        settings = self.serial_port.getSettingsDict()
        settings["timeout"] = 0.1  # seconds
        self.serial_port.applySettingsDict(settings)
        self._ready()
        while not self.finished.is_set():
            # Grab data from the serial port, or wait for timeout if none available.
            raw = self.serial_port.read(max(1, self.serial_port.inWaiting()))

            if not raw:
                # Timeout, finish the current stroke.
                self._finish_stroke()
                continue

            for byte in raw:
                key_set = byte >> 6
                if key_set <= self._last_key_set:
                    # Starting a new stroke, finish previous one.
                    self._finish_stroke()
                self._last_key_set = key_set
                for i in range(5 if key_set == 3 else 6):
                    if (byte >> i) & 1:
                        key = STENO_KEY_CHART[(key_set * 6) + i]
                        self._pressed_keys.append(key)
                if key_set == 3:
                    # Last possible set, the stroke is finished.
                    self._finish_stroke()



# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

"""Centralized place for dictionary loading operation."""

import threading
import time

from plover.dictionary.base import load_dictionary
from plover.resource import resource_timestamp
from plover import log


class DictionaryLoadingManager:
    def __init__(self, state_change_callback):
        """
        Parameters:
        state_change_callback -- A function that will be called when any dictionary is loaded
            with two parameters: the filename and the loaded StenoDictionary object
            (or an instance of ErroredDictionary if the load fails).
        """
        self._state_change_callback = state_change_callback
        self.dictionaries = {}

    def __len__(self):
        return len(self.dictionaries)

    def __getitem__(self, filename):
        return self.dictionaries[filename].get()

    def __contains__(self, filename):
        return filename in self.dictionaries

    def start_loading(self, filename):
        op = self.dictionaries.get(filename)
        if op is not None and not op.needs_reloading():
            return op
        log.info(
            "%s dictionary: %s", "loading" if op is None else "reloading", filename
        )
        op = DictionaryLoadingOperation(filename, self._state_change_callback)
        self.dictionaries[filename] = op
        return op

    def unload_outdated(self):
        for filename, op in list(self.dictionaries.items()):
            if op.needs_reloading():
                del self.dictionaries[filename]

    def load(self, filenames):
        start_time = time.time()
        self.dictionaries = {f: self.start_loading(f) for f in filenames}
        results = [self.dictionaries[f].get() for f in filenames]
        log.info(
            "loaded %u dictionaries in %.3fs", len(results), time.time() - start_time
        )
        return results


class DictionaryLoadingOperation:
    def __init__(self, filename, state_change_callback):
        """
        Parameters:
        filename -- Path to dictionary file.
        state_change_callback -- A function that will be called when the load is finished
            with two parameters: the filename and the loaded StenoDictionary object
            (or an instance of ErroredDictionary if the load fails).
        """
        self._state_change_callback = state_change_callback
        self.loading_thread = threading.Thread(target=self.load)
        self.filename = filename
        self.result = None
        self.loading_thread.start()

    def needs_reloading(self):
        try:
            new_timestamp = resource_timestamp(self.filename)
        except:
            # Bad resource name, permission denied, path
            # does not exist, ...
            new_timestamp = None
        # If no change in timestamp: don't reload.
        if new_timestamp == self.result.timestamp:
            return False
        # If we could not get the new timestamp:
        # the dictionary is not available anymore,
        # attempt to reload to notify the user.
        if new_timestamp is None:
            return True
        # If we're here, and no previous timestamp exists,
        # then the file was previously inaccessible, reload.
        if self.result.timestamp is None:
            return True
        # Otherwise, just compare timestamps.
        return self.result.timestamp < new_timestamp

    def load(self):
        timestamp = None
        try:
            timestamp = resource_timestamp(self.filename)
            self.result = load_dictionary(self.filename)
        except Exception as e:
            log.debug("loading dictionary %s failed", self.filename, exc_info=True)
            from plover.engine import ErroredDictionary

            self.result = ErroredDictionary(self.filename, e)
            self.result.timestamp = timestamp
        self._state_change_callback(self.filename, self.result)

    def get(self):
        self.loading_thread.join()
        return self.result
